# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import torch
import six
import ubelt as ub


class ModuleMixin(object):
    """
    Adds convenience functions to a torch module
    """

    def trainable_layers(self, names=False):
        """
        Get the layers netharn identifies as "trainable"

        Example:
            >>> import torchvision
            >>> model = torchvision.models.AlexNet()
            >>> list(ModuleMixin.trainable_layers(model, names=True))
        """
        return trainable_layers(self, names=names)

    def number_of_parameters(self, trainable=True):
        """
        Tally the number of model paramters.

        Args:
            trainable (bool, default=True): specify if only trainable
                params should be counted.

        Returns:
            int: number of paramaters in this module and all submodules
        """
        return number_of_parameters(self, trainable)

    def _device_dict(self):
        return {key: item.device for key, item in self.state_dict().items()}

    def devices(self):
        """
        Returns all devices this module state is mounted on

        Returns:
            Set[torch.device]: set of devices used by this model

        Example:
            >>> import netharn as nh
            >>> self = nh.models.toynet.ToyNet2d()
            >>> ub.peek(self.devices())
            device(type='cpu')
            >>> # xdoctest: +REQUIRES(--multigpu)
            >>> self = nh.XPU([0, 1]).mount(self)
            >>> print(self.devices())
            {device(type='cuda', index=0), device(type='cuda', index=1)}
            >>> print('self.main_device = {!r}'.format(self.main_device))
            self.main_device = device(type='cuda', index=0)
        """
        state_devices = self._device_dict()
        devices = set(state_devices.values())
        if hasattr(self, 'device_ids'):
            # Handle data parallel
            for _id in self.device_ids:
                devices.add(torch.device(_id))
        return devices

    @property
    def main_device(self):
        """
        The main/src torch device used by this model
        """
        if hasattr(self, 'src_device_obj'):
            return self.src_device_obj
        else:
            devices = self.devices()
            if len(devices) > 1:
                raise NotImplementedError('no information maintained on which device is primary')
            else:
                return list(devices)[0]


def number_of_parameters(model, trainable=True):
    """
    Returns number of trainable parameters in a torch module

    Example:
        >>> import netharn as nh
        >>> model = nh.models.ToyNet2d()
        >>> number_of_parameters(model)
        824
    """
    if trainable:
        model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    else:
        model_parameters = model.parameters()
    n_params = sum([np.prod(p.size()) for p in model_parameters])
    return n_params


class grad_context(object):
    """
    Context manager for controlling if autograd is enabled.

    DEPRECATED use ``torch.set_grad_enabled`` instead
    """
    def __init__(self, flag):
        import warnings
        warnings.warn('Deprecated use torch.set_grad_enabled instead',
                      DeprecationWarning)
        if tuple(map(int, torch.__version__.split('.')[0:2])) < (0, 4):
            self.prev = None
            self.flag = flag
        else:
            self.prev = torch.is_grad_enabled()
            self.flag = flag

    def __enter__(self):
        if self.prev is not None:
            torch.set_grad_enabled(self.flag)

    def __exit__(self, *args):
        if self.prev is not None:
            torch.set_grad_enabled(self.prev)
            return False


def _get_method_func(method):
    func = method.im_func if six.PY2 else method.__func__
    return func


def _get_method_base_class(method):
    """
    Finds the class in which a particular method function was defined.

    CommandLine:
        xdoctest -m netharn.util.util_torch _get_method_base_class

    Example:
        >>> method = torch.nn.BatchNorm2d(1).forward
        >>> print(_get_method_base_class(method))
        <class 'torch.nn.modules.batchnorm._BatchNorm'>
    """
    import sys
    if six.PY2:
        from qualname import qualname as find_qualname
        qualname = find_qualname(method)
    else:
        qualname = method.__qualname__
    module = sys.modules[method.__module__]
    base_name = qualname.split('.')[0]
    base_cls = getattr(module, base_name)
    return base_cls


class IgnoreLayerContext(object):
    """
    Context manager that modifies (monkey-patches) models to temporarily
    remove the forward pass for particular layers.

    Args:
        model (torch.nn.Module): model to modify

        category (type): the module class to be ignored

        enabled (bool, default=True): if True this context manager is enabled
            otherwise it does nothing (i.e. the specified layers will not be
            ignored).

    Example:
        >>> input = torch.rand(1, 1, 10, 10)
        >>> model = torch.nn.BatchNorm2d(1)
        >>> output1 = model(input)
        >>> with IgnoreLayerContext(model, torch.nn.BatchNorm2d):
        ...     output2 = model(input)
        >>> output3 = model(input)
        >>> assert torch.all(output3 == output1)
        >>> assert torch.all(output2 == input)

    Ignore:
        >>> # Test issue with data parallel
        >>> from netharn.util.util_torch import *
        >>> import torch
        >>> import netharn as nh
        >>> layer = raw_model = torch.nn.BatchNorm2d(1)
        >>> raw_inputs = torch.rand(8, 1, 10, 10)
        >>> xpu = nh.XPU.coerce([0,1])
        >>> model = xpu.mount(raw_model)
        >>> inputs = xpu.move(raw_inputs)
        >>> output1 = model(inputs)
        >>> with nh.util.IgnoreLayerContext(model, torch.nn.BatchNorm2d):
        ...     print('model.module.forward = {!r}'.format(model.module.forward))
        ...     output2 = model(inputs)
        >>> output3 = model(inputs)
        >>> assert torch.all(output3 == output1)
        >>> assert torch.all(output2 == inputs)
        >>> # ------------
        >>> raw_model = torch.nn.BatchNorm2d(1)
        >>> raw_inputs = torch.rand(8, 1, 10, 10)
        >>> model = xpu.mount(raw_model)
        >>> inputs = xpu.move(raw_inputs)
        >>> output1 = model(inputs)
        >>> self = nh.util.IgnoreLayerContext(model, torch.nn.BatchNorm2d)
        >>> self.__enter__()
        >>> output2 = model(inputs)
        >>> self.__exit__()
        >>> print('CAN WE DO THIS?')
        >>> output3 = model(inputs)
        >>> # ------------
        >>> xpu = nh.XPU.coerce([0,1])
        >>> devices = [torch.device(type='cuda', index=i) for i in [0, 1]]
        >>> replicas = torch.nn.parallel.replicate(xpu.move(raw_model), devices)
        >>> [r.forward for r in replicas]
        >>> print([r.weight for r in replicas])
        >>> r = replicas[1]
        >>> # ------
        >>> assert torch.all(output3 == output1)
        >>> assert torch.all(output2 == inputs)
    """
    def __init__(self, model, category=None, enabled=True):
        self.model = model
        self.category = category
        self.prev_state = None

        self._PATCH_CLASS = True  # are we patching the instance or class?

    def __enter__(self):
        self.prev_state = {}
        def _noop_forward(self, inputs, *args, **kwargs):
            return inputs
        _noop_forward._patched = True

        for name, layer in trainable_layers(self.model, names=True):
            needs_filter = False
            if self.category is not None:
                needs_filter |= isinstance(layer, self.category)

            if needs_filter:
                if self._PATCH_CLASS:
                    func = _get_method_func(layer.forward)
                    already_patched = not getattr(func, '_patched', False)
                    if already_patched:
                        # Patch the entire class if it wasn't already
                        base_cls = _get_method_base_class(layer.forward)
                        assert 'forward' in base_cls.__dict__
                        # print('PATCH FORWARD IN base_cls = {!r}'.format(base_cls))
                        # print('base_cls = {!r}'.format(base_cls))
                        self.prev_state[name] = (base_cls, base_cls.forward)
                        base_cls.forward = _noop_forward
                else:
                    self.prev_state[name] = layer.forward
                    ub.inject_method(layer, _noop_forward, name='forward')
        return self

    def __exit__(self, *args):
        if self.prev_state:
            if self._PATCH_CLASS:
                # Unpatch all patched classes
                for name, state in self.prev_state.items():
                    base_cls, orig = state
                    base_cls.forward = orig
            else:
                for name, layer in trainable_layers(self.model, names=True):
                    if name in self.prev_state:
                        # Unset the instance attribute that overrides the default
                        # class function attribute. Note that we cannot simply
                        # reset the forward attribute to its old value because that
                        # will still leave an entry in the layer.__dict__ that
                        # previously wasn't there. Having the forward method
                        # populated in layer.__dict__ causes issues with data
                        # parallel.
                        del layer.__dict__['forward']


class BatchNormContext(object):
    """
    Sets batch norm training state of `model` to `training` within the context
    manager.

    Args:
        model (torch.nn.Module | Sequnce[Module]): model(s) to modify

        training (bool, default=False):
            if True training of batch norm layers is enabled otherwise it is
            disabled. This is useful for batches of size 1.
    """
    def __init__(self, models, training=True, **kw):
        if not isinstance(models, (tuple, list)):
            models = [models]
        self.models = models
        if kw:
            import warnings
            warnings.warn('the enabled kwarg is deprecated')
            training = kw.pop('enabled', training)
            if len(kw):
                raise ValueError('Unsupported kwargs: {}'.format(list(kw)))
        self.training = training
        self.prev_train_state = None

    def __enter__(self):
        self.prev_train_state = ub.ddict(dict)
        for i, model in enumerate(self.models):
            for name, layer in trainable_layers(model, names=True):
                if isinstance(layer, torch.nn.modules.batchnorm._BatchNorm):
                    self.prev_train_state[i][name] = layer.training
                    layer.training = self.training
        return self

    def __exit__(self, *args):
        if self.prev_train_state:
            for i, model in enumerate(self.models):
                for name, layer in trainable_layers(model, names=True):
                    if name in self.prev_train_state[i]:
                        layer.training = self.prev_train_state[i][name]


DisableBatchNorm = BatchNormContext


def trainable_layers(model, names=False):
    """
    Note:
        This was moved to netharn.initializers.functional.
        Move it back here, or do some other refactoring.

    Example:
        >>> import torchvision
        >>> model = torchvision.models.AlexNet()
        >>> list(trainable_layers(model, names=True))
    """
    if names:
        stack = [('', '', model)]
        while stack:
            prefix, basename, item = stack.pop()
            name = '.'.join([p for p in [prefix, basename] if p])
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield name, item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield name, item
            elif hasattr(item, 'reset_parameters'):
                yield name, item

            child_prefix = name
            for child_basename, child_item in list(item.named_children())[::-1]:
                stack.append((child_prefix, child_basename, child_item))
    else:
        queue = [model]
        while queue:
            item = queue.pop(0)
            # TODO: need to put all trainable layer types here
            # (I think this is just everything with reset_parameters)
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield item
            elif hasattr(item, 'reset_parameters'):
                yield item
            # if isinstance(input, torch.nn.modules.Linear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Bilinear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Embedding):
            #     yield item
            # if isinstance(input, torch.nn.modules.EmbeddingBag):
            #     yield item
            for child in item.children():
                queue.append(child)


def one_hot_embedding(labels, num_classes, dtype=None):
    """
    Embedding labels to one-hot form.

    Args:
      labels: (LongTensor) class labels, sized [N,].
      num_classes: (int) number of classes.

    Returns:
      (tensor) encoded labels, sized [N,#classes].

    References:
        https://discuss.pytorch.org/t/convert-int-into-one-hot-format/507/4

    CommandLine:
        python -m netharn.loss one_hot_embedding

    Example:
        >>> # each element in target has to have 0 <= value < C
        >>> labels = torch.LongTensor([0, 0, 1, 4, 2, 3])
        >>> num_classes = max(labels) + 1
        >>> t = one_hot_embedding(labels, num_classes)
        >>> assert all(row[y] == 1 for row, y in zip(t.numpy(), labels.numpy()))
        >>> import ubelt as ub
        >>> print(ub.repr2(t.numpy().tolist()))
        [
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0],
        ]
        >>> t2 = one_hot_embedding(labels.numpy(), num_classes)
        >>> assert np.all(t2 == t.numpy())
        >>> if torch.cuda.is_available():
        >>>     t3 = one_hot_embedding(labels.to(0), num_classes)
        >>>     assert np.all(t3.cpu().numpy() == t.numpy())
    """
    if isinstance(labels, np.ndarray):
        dtype = dtype or np.float
        y = np.eye(num_classes, dtype=dtype)
        y_onehot = y[labels]
    else:  # if torch.is_tensor(labels):
        dtype = dtype or torch.float
        y = torch.eye(num_classes, device=labels.device, dtype=dtype)
        y_onehot = y[labels]
    return y_onehot


def one_hot_lookup(probs, labels):
    """
    Return probbility of a particular label (usually true labels) for each item

    Each item in labels corresonds to a row in probs. Returns the index
    specified at each row.

    Example:
        >>> probs = np.array([
        >>>     [0, 1, 2],
        >>>     [3, 4, 5],
        >>>     [6, 7, 8],
        >>>     [9, 10, 11],
        >>> ])
        >>> labels = np.array([0, 1, 2, 1])
        >>> one_hot_lookup(probs, labels)
        array([ 0,  4,  8, 10])
    """
    return probs[np.eye(probs.shape[1], dtype=np.bool)[labels]]


def torch_ravel_multi_index(multi_index, dims=None, device=None, strides_=None):
    """
    Implementation of `numpy.ravel_multi_index` for torch Tensors

    Args:
        multi_index (List[Tensor] | Tensor) : either a list of indices for
           each dimension, or a tensor containing the same information as
           column vectors.
        dims (Tuple): shape of the array to index into before flattening.
        device (torch.device): default torch device
        strides_ (Tensor, optional): prespecify strides for each dimension
            instead of using dims to compute it.

    Returns:
        torch.LongTensor: flat indices

    Example:
        >>> import torch
        >>> N = 10
        >>> B, A, H, W = 2, 3, 3, 3
        >>> dims = [B, A, 4, H, W]
        >>> device = 'cpu'
        >>> multi_index = [(torch.rand(N) * d).long().to(device) for d in dims]
        >>> multi_index_np = [d.data.cpu().numpy() for d in multi_index]
        >>> numpy_result = np.ravel_multi_index(multi_index_np, dims)
        >>> torch_result = torch_ravel_multi_index(multi_index, dims, device)
        >>> assert np.all(torch_result.data.cpu().numpy() == numpy_result)

    Benchmark:
        >>> import torch
        >>> import ubelt as ub
        >>> N = 10000
        >>> B, A, H, W = 2, 3, 3, 3
        >>> dims = [B, A, 4, H, W]
        >>> device = torch.device('cuda')
        >>> multi_index = [(torch.rand(N) * d).long().to(device) for d in dims]
        >>> #
        >>> ti = ub.Timerit(1000, bestof=10, label='time')
        >>> #
        >>> for timer in ti.reset('cuda'):
        >>>     with timer:
        >>>         torch_ravel_multi_index(multi_index, dims, device)
        >>>         torch.cuda.synchronize()
        >>> #
        >>> for timer in ti.reset('numpy-for-gpu'):
        >>>     with timer:
        >>>         multi_index_np = [d.data.cpu().numpy() for d in multi_index]
        >>>         np.ravel_multi_index(multi_index_np, dims)
        >>>         torch.cuda.synchronize()
        >>> #
        >>> multi_index_np = [d.data.cpu().numpy() for d in multi_index]
        >>> for timer in ti.reset('numpy-native'):
        >>>     with timer:
        >>>         np.ravel_multi_index(multi_index_np, dims)
        >>>         torch.cuda.synchronize()
        >>> #
        >>> strides = np.cumprod(dims[::-1])[::-1][1:].tolist() + [1]
        >>> strides_ = torch.LongTensor(strides).to(device)
        >>> for timer in ti.reset('cuda-precomp-strides'):
        >>>     with timer:
        >>>         torch_ravel_multi_index(multi_index, dims, device, strides_)
        >>>         torch.cuda.synchronize()
    """
    if strides_ is None:
        if 1:
            # this one is the fastest, my guess is because torch
            # doesnt have to worry about managing numpy memory
            nextdim = torch.LongTensor(list(dims[1:]) + [1])
            strides_ = torch.flip(torch.flip(nextdim, (0,)).cumprod(dim=0), (0,))
            strides_ = strides_.to(device)
        elif 0:
            strides = np.cumprod(dims[::-1])[::-1][1:].tolist() + [1]
            strides_ = torch.LongTensor(strides).to(device)
        elif 0:
            strides = np.cumprod([1] + list(dims[::-1][0:-1]))[::-1]
            strides = np.ascontiguousarray(strides)
            strides_ = torch.LongTensor(strides).to(device)
        else:
            strides = np.cumprod([1] + list(dims[::-1][0:-1]))[::-1]
            strides = np.ascontiguousarray(strides)
            strides_ = torch.from_numpy(strides).to(device)
            # strides_ = torch.LongTensor(strides.tolist()).to(device)
    if isinstance(multi_index, (list, tuple)):
        multi_index = torch.cat([x.view(-1, 1) for x in multi_index], dim=1)

    # Could do a mm product if that gets implemented for LongTensors
    result = (multi_index * strides_).sum(dim=1).contiguous()

    # if 1:
    #     if 1:
    #         strides_ = torch.LongTensor(strides).to(device)
    #         multi_index_ = torch.cat([x.view(-1, 1) for x in multi_index], dim=1)
    #         result = (multi_index_ * strides_).sum(dim=1)
    #         # [x * s for s, x in zip(strides_, multi_index)]
    #     else:
    #         strides_ = torch.FloatTensor(strides).view(-1, 1).to(device)
    #         multi_index_ = torch.cat([x.view(-1, 1) for x in multi_index], dim=1).float()
    #         result = torch.mm(multi_index_, strides_).view(-1).long()

    #     # result = (multi_index_ * strides_.view(-1)).sum(dim=1)
    # else:
    #     flat_size = len(multi_index[0])
    #     result = torch.zeros(flat_size, dtype=torch.long, device=device)
    #     for stride, index in zip(strides, multi_index):
    #         if len(index) == 1 and isinstance(index, list):
    #             index = torch.LongTensor(index * flat_size).to(device)
    #         result += stride * index

    return result


class freeze_params(object):
    """
    Context manager for freezing / unfreezing specific layers / params.

    Args:
        layers (Module | List[Module]):
            the modules containing the params you want to freeze

        enabled (bool, default=True): if this layer is enabled

    Example:
        >>> inputs, net = freeze_params.demodata()
        >>> self = freeze_params(net)
        >>> list(self.named_parameters())

    Example:
        >>> from netharn.util.util_torch import *
        >>> inputs, net = freeze_params.demodata()
        >>> with freeze_params(net[1]) as self:
        >>>     x = net(inputs)
        >>> loss = x.sum()
        >>> loss.backward()
        >>> for k, v in net.named_parameters():
        >>>     print('{}.grad = {}'.format(k, v.grad))

        >>> inputs, net = freeze_params.demodata()
        >>> with freeze_params([net[1], net[2]]) as self:
        >>>     x = net(inputs)
        >>> p1 = net(inputs).sum()
        >>> p2 = x.sum()
        >>> loss = p1 + p2
        >>> loss.backward()
        >>> for k, v in net.named_parameters():
        >>>     print('{}.grad = {}'.format(k, v.grad))

        >>> inputs, net = freeze_params.demodata()
        >>> with freeze_params([net[1], net[2]]) as self:
        >>>     x = net(inputs)
        >>> loss = x.sum()
        >>> loss.backward()
        >>> for k, v in net.named_parameters():
        >>>     print('{}.grad = {}'.format(k, v.grad))
    """

    def __init__(self, layers, enabled=True):
        if isinstance(layers, torch.nn.Module):
            self.layers = [layers]
        else:
            if ub.iterable(layers):
                self.layers = list(layers)
            else:
                raise TypeError(layers)
        self.enabled = enabled
        self.state = {}

    @classmethod
    def demodata(cls):
        import netharn as nh
        inputs = torch.rand(2, 1, 1, 1)
        net = nh.layers.Sequential(
            nh.layers.ConvNorm2d(1, 1, 1),
            nh.layers.ConvNorm2d(1, 1, 1),
            nh.layers.ConvNorm2d(1, 1, 1),
        )
        return inputs, net

    def named_parameters(self):
        """ Iterate through all the parameters this context manages """
        for i, layer in enumerate(self.layers):
            prefix = str(i)
            for name, param in layer.named_parameters(prefix, recurse=True):
                yield name, param

    def _set_requires_grad(self, flag):
        for name, param in self.named_parameters():
            param.requires_grad = flag

    def _push_state(self):
        assert len(self.state) == 0, 'can only save ones state'
        for name, param in self.named_parameters():
            self.state[name] = param.requires_grad

    def _pop_state(self):
        for name, param in self.named_parameters():
            param.requires_grad = self.state.pop(name)

    def __enter__(self):
        if self.enabled:
            self._push_state()
            self._set_requires_grad(False)
        return self

    def __exit__(self, a, b, c):
        if self.enabled:
            self._pop_state()



# -*- coding: utf-8 -*-
"""
fit_harness takes your hyperparams and
applys standardized "state-of-the-art" training procedures

But everything is overwritable.
Experimentation and freedom to protype quickly is extremely important.
We do our best not to get in the way, just performing a jumping off point.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub
import torch
import torch.nn
import torchvision  # NOQA
import torch.nn.functional as F
from torch import nn
from os.path import join
import netharn as nh
import numpy as np


class MnistNet(nn.Module):
    def __init__(self, classes, num_channels=1):
        super(MnistNet, self).__init__()
        self.classes = classes
        self.conv1 = nn.Conv2d(num_channels, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, len(self.classes))

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


class MnistHarn(nh.FitHarn):
    """
    Customize relevant parts of the training loop
    """

    def prepare_batch(harn, raw_batch):
        """
        Ensure batch is in a standardized structure
        """
        # Simply move the data from the datasets on the the GPU(s)
        inputs, labels = raw_batch
        inputs = harn.xpu.move(inputs)
        labels = harn.xpu.move(labels)
        batch = {
            'input': inputs,
            'label': labels,
        }
        return batch

    def run_batch(harn, batch):
        """ Core learning / backprop """
        inputs = batch['input']
        labels = batch['label']

        outputs = harn.model(inputs)

        loss = harn.criterion(outputs, labels)
        return outputs, loss

    def on_batch(harn, batch, outputs, loss):
        """ Compute relevent metrics to monitor """
        class_probs = torch.nn.functional.softmax(outputs, dim=1)
        scores, pred = class_probs.max(dim=1)

        pred_labels = pred.cpu().numpy()
        true_labels = batch['label'].cpu().numpy()

        if harn.batch_index < 3:
            import kwimage
            decoded = harn._decode(outputs, batch['label'])
            stacked = harn._draw_batch(batch, decoded)
            dpath = ub.ensuredir((harn.train_dpath, 'monitor', harn.current_tag))
            fpath = join(dpath, 'epoch_{}_batch_{}.jpg'.format(harn.epoch, harn.batch_index))
            kwimage.imwrite(fpath, stacked)

        acc = (true_labels == pred_labels).mean()

        metrics_dict = {
            'acc': acc,
        }
        return metrics_dict

    def _decode(harn, outputs, true_cxs=None):
        class_probs = torch.nn.functional.softmax(outputs, dim=1)
        pred_scores, pred_cxs = class_probs.data.max(dim=1)

        decoded = {
            'class_probs': class_probs,
            'pred_cxs': pred_cxs,
            'pred_scores': pred_scores,
        }
        if true_cxs is not None:
            import kwarray
            hot = kwarray.one_hot_embedding(true_cxs, class_probs.shape[1])
            true_probs = (hot * class_probs).sum(dim=1)
            decoded['true_scores'] = true_probs
        return decoded

    def _draw_batch(harn, batch, decoded, limit=32):
        """
        Example:
            >>> # xdoctest: +REQUIRES(--download)
            >>> harn = setup_harn().initialize()
            >>> #
            >>> batch = harn._demo_batch(0, tag='test')
            >>> outputs, loss = harn.run_batch(batch)
            >>> bx = harn.bxs[harn.current_tag]
            >>> decoded = harn._decode(outputs, batch['label'])
            >>> fpath = harn._draw_batch(bx, batch, decoded, limit=42)
            >>> print('fpath = {!r}'.format(fpath))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(fpath, colorspace='rgb', doclf=True)
            >>> kwplot.show_if_requested()
        """
        import kwimage
        import kwplot
        inputs = batch['input']
        inputs = inputs[0:limit]

        input_shape = inputs.shape
        dims = [160] * (len(input_shape) - 2)
        min_, max_ = inputs.min(), inputs.max()
        inputs = (inputs - min_) / (max_ - min_)
        inputs = torch.nn.functional.interpolate(inputs, size=dims)
        inputs = (inputs * 255).byte()
        inputs = inputs.data.cpu().numpy()

        dset = harn.datasets[harn.current_tag]

        true_cxs = batch['label'].data.cpu().numpy()
        pred_cxs = decoded['pred_cxs'].data.cpu().numpy()
        class_probs = decoded['class_probs'].data.cpu().numpy()

        todraw = []
        for im, pcx, tcx, probs in zip(inputs, pred_cxs, true_cxs, class_probs):
            im_ = im.transpose(1, 2, 0)
            im_ = kwimage.convert_colorspace(im_, 'gray', 'rgb')
            im_ = np.ascontiguousarray(im_)
            im_ = kwplot.draw_clf_on_image(im_, dset.classes, tcx, probs)
            todraw.append(im_)

        stacked = kwimage.stack_images_grid(todraw, overlap=-10, bg_value=(10, 40, 30), chunksize=8)
        return stacked


def setup_datasets(workdir=None):
    if workdir is None:
        workdir = ub.expandpath('~/data/mnist/')

    # Define your dataset
    transform = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize((0.1307,), (0.3081,))
    ])

    learn_dset = torchvision.datasets.MNIST(workdir, transform=transform,
                                            train=True, download=True)

    test_dset = torchvision.datasets.MNIST(workdir, transform=transform,
                                           train=False, download=True)

    # split the learning dataset into training and validation
    # take a subset of data
    factor = .15
    n_vali = int(len(learn_dset) * factor)
    learn_idx = np.arange(len(learn_dset))

    rng = np.random.RandomState(0)
    rng.shuffle(learn_idx)

    reduction = int(ub.argval('--reduction', default=1))
    vali_idx  = torch.LongTensor(learn_idx[:n_vali][::reduction])
    train_idx = torch.LongTensor(learn_idx[n_vali:][::reduction])

    train_dset = torch.utils.data.Subset(learn_dset, train_idx)
    vali_dset = torch.utils.data.Subset(learn_dset, vali_idx)

    classes = ['zero', 'one', 'two', 'three', 'four', 'five', 'six', 'seven',
               'eight', 'nine']

    datasets = {
        'train': train_dset,
        'vali': vali_dset,
        'test': test_dset,
    }
    for tag, dset in datasets.items():
        dset.classes = classes
        dset.num_classes = len(classes)

    # Give the training dataset an input_id
    datasets['train'].input_id = 'mnist_' + ub.hash_data(train_idx.numpy())[0:8]
    return datasets, workdir


def setup_harn(**kw):
    """
    CommandLine:
        python examples/mnist.py

        python ~/code/netharn/examples/mnist.py --gpu=2
        python ~/code/netharn/examples/mnist.py
    """
    xpu = nh.XPU.from_argv(min_memory=300)

    config = {
        'batch_size': kw.get('batch_size', 128),
        'workers': kw.get('workers', 6 if xpu.is_gpu() else 0)
    }

    if config['workers'] > 0:
        # Workaround deadlocks with DataLoader
        import cv2
        cv2.setNumThreads(0)

    datasets, workdir = setup_datasets()

    loaders = {
        tag: torch.utils.data.DataLoader(
            dset, batch_size=config['batch_size'],
            num_workers=config['workers'],
            shuffle=(tag == 'train'))
        for tag, dset in datasets.items()
    }

    if False:
        initializer = (nh.initializers.Pretrained, {
            'fpath': 'path/to/pretained/weights.pt'
        })
    else:
        initializer = (nh.initializers.KaimingNormal, {})

    # Here is the FitHarn magic.
    # They nh.HyperParams object keeps track of and helps log all declarative
    # info related to training a model.
    hyper = nh.hyperparams.HyperParams(
        nice='my-mnist-demo',
        xpu=xpu,
        workdir=workdir,
        datasets=datasets,
        loaders=loaders,
        model=(MnistNet, dict(num_channels=1, classes=datasets['train'].classes)),
        # optimizer=torch.optim.AdamW,
        optimizer=(torch.optim.SGD, {'lr': 0.01, 'weight_decay': 3e-6}),
        # scheduler='ReduceLROnPlateau',
        scheduler=(nh.schedulers.ListedScheduler, {
            'points': {
                'lr': {
                    0   : 0.01,
                    2   : 0.05,
                    10  : 0.10,
                    20  : 0.01,
                    40  : 0.0001,
                },
                'momentum': {
                    0   : 0.95,
                    10  : 0.85,
                    20  : 0.95,
                    40  : 0.99,
                },
                'weight_decay': {
                    0: 3e-6,
                }
            },
            'interpolation': 'linear',
        }),
        criterion=torch.nn.CrossEntropyLoss,
        initializer=initializer,
        monitor=(nh.Monitor, {
            'minimize': ['loss'],
            'maximize': ['acc'],
            'patience': 10,
            'max_epoch': 300,
            'smoothing': .4,
        }),
    )

    harn = MnistHarn(hyper=hyper)
    harn.preferences.update({
        'keyboard_debug': True,
    })

    # Set how often vali / test will be run
    harn.intervals.update({
        # 'vali': slice(5, None, 1),
        # Start testing after the 5th epoch and then test every 4 epochs
        'test': slice(5, None, 4),
    })
    return harn


def main():
    harn = setup_harn()
    reset = ub.argflag('--reset')

    # Initializing a FitHarn object can take a little time, but not too much.
    # This is where instances of the model, optimizer, scheduler, monitor, and
    # initializer are created. This is also where we check if there is a
    # pre-existing checkpoint that we can restart from.
    harn.initialize(reset=reset)

    if ub.argflag(('--vd', '--view-directory')):
        ub.startfile(harn.train_dpath)

    # This starts the main loop which will run until a the monitor's terminator
    # criterion is satisfied. If the initialize step loaded a checkpointed that
    # already met the termination criterion, then this will simply return.
    deploy_fpath = harn.run()

    # The returned deploy_fpath is the path to an exported netharn model.
    # This model is the on with the best weights according to the monitor.
    print('deploy_fpath = {!r}'.format(deploy_fpath))


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.examples.mnist

        tensorboard --logdir ~/data/work/mnist/fit/nice
    """
    main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import ubelt as ub


def compact_idstr(dict_):
    """
    A short unique id string for a dict param config that is semi-interpretable
    """
    from netharn import util
    import ubelt as ub
    short_keys = util.shortest_unique_prefixes(dict_.keys())
    short_dict = ub.odict(sorted(zip(short_keys, dict_.values())))
    idstr = ub.repr2(short_dict, nobr=1, itemsep='', si=1, nl=0,
                     explicit=1)
    return idstr


def make_idstr(d):
    """
    Make full-length-key id-string
    """
    if d is None:
        return ''
    elif isinstance(d, six.string_types):
        return d
    elif len(d) == 0:
        return ''
    if not isinstance(d, ub.odict):
        d = ub.odict(sorted(d.items()))
    return ub.repr2(d, itemsep='', nobr=True, explicit=True, nl=0, si=True)


def make_short_idstr(params, precision=None):
    """
    Make id-string where they keys are shortened

    Args:
        params (dict):

    Returns:
        str:

    CommandLine:
        python -m netharn.util.misc make_short_idstr

    Example:
        >>> # xdoctest: +SKIP
        >>> params = {'input_shape': (None, 3, 212, 212),
        >>>           'a': 'b',
        >>>           'center': {'im_mean': .5, 'std': 1},
        >>>           'alphabet': 'abc'}
        >>> print(make_short_idstr(params))
    """
    if params is None:
        return ''
    elif len(params) == 0:
        return ''
    from netharn import util
    short_keys = util.shortest_unique_prefixes(list(params.keys()),
                                               allow_simple=False,
                                               allow_end=True,
                                               min_length=1)
    def shortval(v):
        if isinstance(v, bool):
            return int(v)
        return v
    d = dict(zip(short_keys, map(shortval, params.values())))
    def make_idstr(d):
        # Note: we are not using sort=True, because repr2 sorts sets and dicts
        # by default.
        return ub.repr2(d, itemsep='', nobr=True, explicit=True, nl=0, si=True,
                        precision=precision).replace(' ', '').replace('[', '').replace(']', '').replace('(', '').replace(')', '').replace('{', '').replace('}', '')
    short_idstr = make_idstr(d)
    return short_idstr



# -*- coding: utf-8 -*-
"""
DEPRECATED
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np


def read_h5arr(fpath):
    import h5py
    with h5py.File(fpath, 'r') as hf:
        return hf['arr_0'][...]


def write_h5arr(fpath, arr):
    import h5py
    with h5py.File(fpath, 'w') as hf:
        hf.create_dataset('arr_0', data=arr)


def read_arr(fpath):
    """
    Example:
        >>> import ubelt as ub
        >>> import netharn as nh
        >>> from os.path import join
        >>> dpath = ub.ensure_app_cache_dir('netharn', 'tests')
        >>> arr = np.random.rand(10)
        >>> fpath = join(dpath, 'arr.npy')
        >>> nh.util.write_arr(fpath, arr)
        >>> arr2 = nh.util.read_arr(fpath)
        >>> assert np.all(arr == arr2)
        >>> # xdoctest: +REQUIRES(module:h5py)
        >>> fpath = join(dpath, 'arr.h5')
        >>> nh.util.write_arr(fpath, arr)
        >>> arr2 = nh.util.read_arr(fpath)
        >>> assert np.all(arr == arr2)
    """
    if fpath.endswith('.npy'):
        return np.load(fpath)
    elif fpath.endswith('.h5'):
        return read_h5arr(fpath)
    else:
        raise KeyError(fpath)


def write_arr(fpath, arr):
    if fpath.endswith('.npy'):
        return np.save(fpath, arr)
    elif fpath.endswith('.h5'):
        return write_h5arr(fpath, arr)
    else:
        raise KeyError(fpath)



"""

mkinit ~/code/kwarray/kwarray/__init__.py --relative --nomods -w
mkinit ~/code/kwimage/kwimage/__init__.py --relative --nomods -w
mkinit ~/code/kwplot/kwplot/__init__.py --relative --nomods -w
mkinit netharn.util --relative --nomods -w
"""
from __future__ import absolute_import, division, print_function, unicode_literals

__extra_all__ = [
    'profiler',
    'util_dataframe',
    'dataframe_light',
]

__private__ = ['util_dataframe']


__external__ = [
    'kwarray', 'kwimage', 'kwplot'
]


# <AUTOGEN_INIT>
from .imutil import (adjust_gamma, ensure_grayscale, get_num_channels,
                     image_slices, load_image_paths, overlay_colorized,
                     wide_strides_1d,)
from .mplutil import (adjust_subplots, aggensure, axes_extent, colorbar,
                      colorbar_image, copy_figure_to_clipboard, draw_border,
                      extract_axes_extents, interpolated_colormap,
                      make_legend_img, pandas_plot_matrix, qtensure,
                      render_figure_to_image, reverse_colormap, save_parts,
                      savefig2, scores_to_cmap, scores_to_color,)
from .profiler import (IS_PROFILING, profile, profile_now,)
from .util_averages import (CumMovingAve, ExpMovingAve, InternalRunningStats,
                            MovingAve, RunningStats, WindowedMovingAve,
                            absdev,)
from .util_cachestamp import (CacheStamp,)
from .util_filesys import (get_file_info,)
from .util_fname import (align_paths, check_aligned, dumpsafe,
                         shortest_unique_prefixes, shortest_unique_suffixes,)
from .util_idstr import (compact_idstr, make_idstr, make_short_idstr,)
from .util_inspect import (default_kwargs,)
from .util_io import (read_arr, read_h5arr, write_arr, write_h5arr,)
from .util_iter import (roundrobin,)
from .util_json import (LossyJSONEncoder, NumpyEncoder,
                        ensure_json_serializable, read_json, walk_json,
                        write_json,)
from .util_misc import (FlatIndexer, SupressPrint, align, align_lines,
                        strip_ansi,)
from .util_resources import (ensure_ulimit,)
from .util_slider import (SlidingWindow, Stitcher,)
from .util_slider_dep import (SlidingIndexDataset, SlidingSlices,)
from .util_subextreme import (argsubmax, argsubmaxima,)
from .util_tensorboard import (read_tensorboard_scalars,)
from .util_torch import (BatchNormContext, DisableBatchNorm,
                         IgnoreLayerContext, ModuleMixin, freeze_params,
                         grad_context, number_of_parameters, one_hot_embedding,
                         one_hot_lookup, torch_ravel_multi_index,
                         trainable_layers,)
from .util_zip import (split_archive, zopen,)
from kwarray import (ArrayAPI, DataFrameArray, DataFrameLight, LocLight,
                     apply_grouping, arglexmax, argmaxima, argminima,
                     atleast_nd, boolmask, ensure_rng, group_consecutive,
                     group_consecutive_indices, group_indices, group_items,
                     isect_flags, iter_reduce_ufunc, maxvalue_assignment,
                     mincost_assignment, mindist_assignment, one_hot_embedding,
                     one_hot_lookup, random_combinations, random_product,
                     seed_global, shuffle, standard_normal, standard_normal32,
                     standard_normal64, stats_dict, uniform, uniform32,)
from kwimage import (BASE_COLORS, Boxes, CSS4_COLORS, Color, Coords,
                     Detections, Heatmap, Mask, MaskList, MultiPolygon, Points,
                     PointsList, Polygon, PolygonList, Segmentation,
                     SegmentationList, TABLEAU_COLORS,
                     TORCH_GRID_SAMPLE_HAS_ALIGN, XKCD_COLORS,
                     atleast_3channels, available_nms_impls,
                     convert_colorspace, daq_spatial_nms, decode_run_length,
                     draw_boxes_on_image, draw_clf_on_image,
                     draw_text_on_image, draw_vector_field, encode_run_length,
                     ensure_alpha_channel, ensure_float01, ensure_uint255,
                     fourier_mask, gaussian_patch, grab_test_image,
                     grab_test_image_fpath, imread, imresize, imscale, imwrite,
                     load_image_shape, make_channels_comparable, make_heatmask,
                     make_orimask, make_vector_field, non_max_supression,
                     num_channels, overlay_alpha_images, overlay_alpha_layers,
                     radial_fourier_mask, rle_translate, smooth_prob,
                     stack_images, stack_images_grid, subpixel_accum,
                     subpixel_align, subpixel_getvalue, subpixel_maximum,
                     subpixel_minimum, subpixel_set, subpixel_setvalue,
                     subpixel_slice, subpixel_translate, warp_points,
                     warp_tensor,)
from kwplot import (Color, PlotNums, autompl, autoplt, distinct_colors,
                    distinct_markers, draw_boxes, draw_boxes_on_image,
                    draw_clf_on_image, draw_line_segments, draw_text_on_image,
                    ensure_fnum, figure, imshow, legend, make_conv_images,
                    make_heatmask, make_orimask, make_vector_field, multi_plot,
                    next_fnum, plot_convolutional_features, plot_matrix,
                    plot_surface3d, set_figtitle, set_mpl_backend,
                    show_if_requested,)

__all__ = ['ArrayAPI', 'BASE_COLORS', 'BatchNormContext', 'Boxes',
           'CSS4_COLORS', 'CacheStamp', 'Color', 'Color', 'Coords',
           'CumMovingAve', 'DataFrameArray', 'DataFrameLight', 'Detections',
           'DisableBatchNorm', 'ExpMovingAve', 'FlatIndexer', 'Heatmap',
           'IS_PROFILING', 'IgnoreLayerContext', 'InternalRunningStats',
           'LocLight', 'LossyJSONEncoder', 'Mask', 'MaskList',
           'ModuleMixin', 'MovingAve', 'MultiPolygon', 'NumpyEncoder',
           'PlotNums', 'Points', 'PointsList', 'Polygon', 'PolygonList',
           'RunningStats', 'Segmentation', 'SegmentationList',
           'SlidingIndexDataset', 'SlidingSlices', 'SlidingWindow', 'Stitcher',
           'SupressPrint', 'TABLEAU_COLORS', 'TORCH_GRID_SAMPLE_HAS_ALIGN',
           'WindowedMovingAve', 'XKCD_COLORS', 'absdev', 'adjust_gamma',
           'adjust_subplots', 'aggensure', 'align', 'align_lines',
           'align_paths', 'apply_grouping', 'arglexmax', 'argmaxima',
           'argminima', 'argsubmax', 'argsubmaxima', 'atleast_3channels',
           'atleast_nd', 'autompl', 'autoplt', 'available_nms_impls',
           'axes_extent', 'boolmask', 'check_aligned', 'colorbar',
           'colorbar_image', 'compact_idstr', 'convert_colorspace',
           'copy_figure_to_clipboard', 'daq_spatial_nms', 'dataframe_light',
           'decode_run_length', 'default_kwargs', 'distinct_colors',
           'distinct_markers', 'draw_border', 'draw_boxes',
           'draw_boxes_on_image', 'draw_boxes_on_image', 'draw_clf_on_image',
           'draw_clf_on_image', 'draw_line_segments', 'draw_text_on_image',
           'draw_text_on_image', 'draw_vector_field', 'dumpsafe',
           'encode_run_length', 'ensure_alpha_channel', 'ensure_float01',
           'ensure_fnum', 'ensure_grayscale', 'ensure_json_serializable',
           'ensure_rng', 'ensure_uint255', 'ensure_ulimit',
           'extract_axes_extents', 'figure', 'fourier_mask', 'freeze_params',
           'gaussian_patch', 'get_file_info', 'get_num_channels',
           'grab_test_image', 'grab_test_image_fpath', 'grad_context',
           'group_consecutive', 'group_consecutive_indices', 'group_indices',
           'group_items', 'image_slices', 'imread', 'imresize', 'imscale',
           'imshow', 'imwrite', 'interpolated_colormap', 'isect_flags',
           'iter_reduce_ufunc', 'legend', 'load_image_paths',
           'load_image_shape', 'make_channels_comparable', 'make_conv_images',
           'make_heatmask', 'make_heatmask', 'make_idstr', 'make_legend_img',
           'make_orimask', 'make_orimask', 'make_short_idstr',
           'make_vector_field', 'make_vector_field', 'maxvalue_assignment',
           'mincost_assignment', 'mindist_assignment', 'multi_plot',
           'next_fnum', 'non_max_supression', 'num_channels',
           'number_of_parameters', 'one_hot_embedding', 'one_hot_embedding',
           'one_hot_lookup', 'one_hot_lookup', 'overlay_alpha_images',
           'overlay_alpha_layers', 'overlay_colorized', 'pandas_plot_matrix',
           'plot_convolutional_features', 'plot_matrix', 'plot_surface3d',
           'profile', 'profile_now', 'profiler', 'qtensure',
           'radial_fourier_mask', 'random_combinations', 'random_product',
           'read_arr', 'read_h5arr', 'read_json', 'read_tensorboard_scalars',
           'render_figure_to_image', 'reverse_colormap', 'rle_translate',
           'roundrobin', 'save_parts', 'savefig2', 'scores_to_cmap',
           'scores_to_color', 'seed_global', 'set_figtitle', 'set_mpl_backend',
           'shortest_unique_prefixes', 'shortest_unique_suffixes',
           'show_if_requested', 'shuffle', 'smooth_prob', 'split_archive',
           'stack_images', 'stack_images_grid', 'standard_normal',
           'standard_normal32', 'standard_normal64', 'stats_dict',
           'strip_ansi', 'subpixel_accum', 'subpixel_align',
           'subpixel_getvalue', 'subpixel_maximum', 'subpixel_minimum',
           'subpixel_set', 'subpixel_setvalue', 'subpixel_slice',
           'subpixel_translate', 'torch_ravel_multi_index', 'trainable_layers',
           'uniform', 'uniform32', 'util_dataframe', 'walk_json',
           'warp_points', 'warp_tensor', 'wide_strides_1d', 'write_arr',
           'write_h5arr', 'write_json', 'zopen']
# </AUTOGEN_INIT>


# Backwards compatibility patch
from . import util_dataframe
from kwarray import dataframe_light



from netharn import util
from torch import nn
import torch


# TODO: move module mixin from util to here
ModuleMixin = util.ModuleMixin


class Module(nn.Module, util.ModuleMixin):
    """
    Like torch.nn.Module but contains ModuleMixin (for number_of_parameters)
    """

    def output_shape_for(self, input_shape):
        raise NotImplementedError

    def receptive_field_for(self, input_field=None):
        raise NotImplementedError


class Sequential(nn.Sequential, util.ModuleMixin):
    """
    Like torch.nn.Sequential but implements output_shape_for

    Example:
        >>> import torch
        >>> import netharn as nh
        >>> import ubelt as ub
        >>> self = nh.layers.Sequential(
        >>>     torch.nn.Conv2d(2, 3, kernel_size=3),
        >>>     torch.nn.Conv2d(3, 5, kernel_size=3),
        >>>     torch.nn.Conv2d(5, 7, kernel_size=3),
        >>> )
        >>> shape = self.output_shape_for([1, 1, 7, 11])
        >>> print('shape = {}'.format(shape))
        >>> print('shape.hidden = {}'.format(ub.repr2(shape.hidden, nl=1)))
        shape = (1, 7, 1, 5)
        shape.hidden = {
            '0': (1, 3, 5, 9),
            '1': (1, 5, 3, 7),
            '2': (1, 7, 1, 5),
        }
    """
    def output_shape_for(self, input_shape):
        from netharn.analytic import output_shape_for
        return output_shape_for.OutputShapeFor.sequential(self, input_shape)

    def receptive_field_for(self, input_field=None):
        from netharn.analytic import receptive_field_for
        return receptive_field_for.ReceptiveFieldFor.sequential(self, input_field)


class Identity(Sequential):
    """
    A identity-function layer.

    Example:
        >>> import torch
        >>> self = Identity()
        >>> a = torch.rand(3, 3)
        >>> b = self(a)
        >>> assert torch.all(a == b)
    """
    def __init__(self):
        super(Identity, self).__init__()

    def output_shape_for(self, input_shape):
        return input_shape

    def receptive_field_for(self, input_field=None):
        if input_field is None:
            from netharn.analytic import receptive_field_for
            input_field = receptive_field_for.ReceptiveFieldFor.input()
        return input_field


class AnalyticModule(Module):
    """
    Provides default implementations for output_shape_for and
    receptive_field_for if _analytic_forward is defined
    """

    def _analytic_forward(self, inputs, _OutputFor, _Output, _Hidden,
                          **kwargs):
        """
        Defines a symbolic representation of forward, output shape, and receptive field.

        Example:
            >>> import netharn as nh
            >>> globals().update(nh.layers.AnalyticModule._analytic_shape_kw())
        """
        # _Output.coerce(output, hidden)
        raise NotImplementedError

    @classmethod
    def _analytic_shape_kw(self):
        from netharn.analytic import output_shape_for
        return {
            '_OutputFor': output_shape_for.OutputShapeFor,
            '_Output': output_shape_for.OutputShape,
            '_Hidden': output_shape_for.HiddenShapes
        }

    @classmethod
    def _analytic_field_kw(self):
        from netharn.analytic import receptive_field_for
        # import netharn as nh
        return {
            '_OutputFor': receptive_field_for.ReceptiveFieldFor,
            '_Output': receptive_field_for.ReceptiveField,
            '_Hidden': receptive_field_for.HiddenFields
        }

    @classmethod
    def _analytic_forward_kw(self):
        # import netharn as nh
        from netharn.analytic import analytic_for
        return {
            '_OutputFor': analytic_for.ForwardFor,
            '_Output': analytic_for.Output,
            '_Hidden': analytic_for.Hidden,
        }

    def output_shape_for(self, input_shape, **kwargs):
        """
        Uses custom _analytic_forward to compute output shape
        """
        kw = self._analytic_shape_kw()
        if kwargs:
            kw = kw.copy()
            kw.update(kwargs)
        return self._analytic_forward(input_shape, **kw)

    def receptive_field_for(self, input_field=None, **kwargs):
        """
        Uses custom _analytic_forward to compute receptive field
        """
        # import netharn as nh
        from netharn.analytic import receptive_field_for
        if input_field is None:
            input_field = receptive_field_for.ReceptiveFieldFor.input()
        kw = self._analytic_field_kw()
        if kwargs:
            kw = kw.copy()
            kw.update(kwargs)
        return self._analytic_forward(input_field, **kw)

    def forward(self, inputs, **kwargs):
        """
        Uses custom _analytic_forward to compute receptive field
        """
        kw = self._analytic_forward_kw()
        if kwargs:
            kw = kw.copy()
            kw.update(kwargs)
        return self._analytic_forward(inputs, **kw)


class Loss(nn.modules.loss._Loss):
    """
    Helper to keep track of if a loss module is in cpu or gpu mod
    """

    def __init__(self):
        super(Loss, self).__init__()
        self._iscuda = False
        self._device_num = None

    def cuda(self, device_num=None, **kwargs):
        self._iscuda = True
        self._device_num = device_num
        return super(Loss, self).cuda(device_num, **kwargs)

    def cpu(self):
        self._iscuda = False
        self._device_num = None
        return super(Loss, self).cpu()

    @property
    def is_cuda(self):
        return self._iscuda

    def get_device(self):
        if self._device_num is None:
            return torch.device('cpu')
        return self._device_num



"""
How do we make a LR scheduler that works at the iteration level instead of the
epoch level? Is this even a good idea?


References:
    # Shows how PJR divides learning rate by batch size
    https://github.com/pjreddie/darknet/blob/e7405b513dc69f17e9e75e8d306d22f2e08c1489/src/convolutional_layer.c#L538

    # Shows how burn_in LR is computed
    https://github.com/pjreddie/darknet/blob/d3828827e70b293a3045a1eb80bfb4026095b87b/src/network.c#L95

    if (batch_num < net->burn_in) return net->learning_rate * pow((float)batch_num / net->burn_in, net->power);


base_lr = .01

burn_in = 1000
power = 4

base_lr * pow(batch_num / burn_in, power)


epoch_frac = np.linspace(0, 1, 10)

base_lr * pow(epoch_frac, power)


"""



import ubelt as ub
from os.path import exists
from os.path import join
from os.path import dirname
from os.path import relpath


def convert_voc_to_coco():
    # TODO: convert segmentation information

    classes = [
        'aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus', 'car', 'cat',
        'chair', 'cow', 'diningtable', 'dog', 'horse', 'motorbike', 'person',
        'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor']
    devkit_dpath = ensure_voc_data()
    root = out_dpath = dirname(devkit_dpath)

    dsets = []

    d = _convert_voc_split(devkit_dpath, classes, 'train', 2012, root)
    dsets.append(d)

    d = _convert_voc_split(devkit_dpath, classes, 'train', 2007, root)
    dsets.append(d)

    d = _convert_voc_split(devkit_dpath, classes, 'val', 2012, root)
    dsets.append(d)

    d = _convert_voc_split(devkit_dpath, classes, 'val', 2007, root)
    dsets.append(d)

    d = _convert_voc_split(devkit_dpath, classes, 'test', 2007, root)
    dsets.append(d)

    if 0:
        import xdev
        xdev.view_directory(out_dpath)

    def reroot_imgs(dset, root):
        for img in dset.imgs.values():
            img['file_name'] = relpath(img['file_name'], root)

    import ndsampler
    t1 = ndsampler.CocoDataset(join(out_dpath, 'voc-train-2007.mscoco.json'))
    t2 = ndsampler.CocoDataset(join(out_dpath, 'voc-train-2012.mscoco.json'))

    v1 = ndsampler.CocoDataset(join(out_dpath, 'voc-val-2007.mscoco.json'))
    v2 = ndsampler.CocoDataset(join(out_dpath, 'voc-val-2012.mscoco.json'))

    t = ndsampler.CocoDataset.union(t1, t2)
    t.tag = 'voc-train'
    t.fpath = join(root, t.tag + '.mscoco.json')

    v = ndsampler.CocoDataset.union(v1, v2)
    v.tag = 'voc-val'
    v.fpath = join(root, v.tag + '.mscoco.json')

    tv = ndsampler.CocoDataset.union(t1, t2, v1, v2)
    tv.tag = 'voc-trainval'
    tv.fpath = join(root, tv.tag + '.mscoco.json')

    print('t.fpath = {!r}'.format(t.fpath))
    t.dump(t.fpath, newlines=True)
    print('v.fpath = {!r}'.format(v.fpath))
    v.dump(v.fpath, newlines=True)
    print('tv.fpath = {!r}'.format(tv.fpath))
    tv.dump(tv.fpath, newlines=True)
    if 0:
        tv.img_root = root
        import kwplot
        kwplot.autompl()
        tv.show_image(2)

    dsets = {
        'train': t,
        'vali': v,
        'trainval': tv,
    }
    return dsets


def _convert_voc_split(devkit_dpath, classes, split, year, root):
    """
    split, year = 'train', 2012
    split, year = 'train', 2007
    """
    import ndsampler
    import xml.etree.ElementTree as ET
    dset = ndsampler.CocoDataset(tag='voc-{}-{}'.format(split, year))

    for catname in classes:
        dset.add_category(catname)

    gpaths, apaths = _read_split_paths(devkit_dpath, split, year)

    KNOWN = {'object', 'segmented', 'size', 'source', 'filename', 'folder',
             'owner'}

    for gpath, apath in ub.ProgIter(zip(gpaths, apaths)):
        tree = ET.parse(apath)
        troot = tree.getroot()

        top_level = list(troot)

        unknown = {e.tag for e in top_level} - KNOWN
        assert not unknown

        img = {
            'file_name': relpath(gpath, root),
            'width': int(tree.find('size').find('width').text),
            'height': int(tree.find('size').find('height').text),
            'depth': int(tree.find('size').find('depth').text),
            'segmented': int(tree.find('segmented').text),
            'source': {
                elem.tag: elem.text
                for elem in list(tree.find('source'))
            },
        }

        assert img.pop('depth') == 3

        owner = tree.find('owner')
        if owner is not None:
            img['owner'] = {
                elem.tag: elem.text
                for elem in list(owner)
            }

        gid = dset.add_image(**img)

        for obj in tree.findall('object'):
            bbox = obj.find('bndbox')
            # Make pixel indexes 0-based
            x1 = float(bbox.find('xmin').text) - 1
            y1 = float(bbox.find('ymin').text) - 1
            x2 = float(bbox.find('xmax').text) - 1
            y2 = float(bbox.find('ymax').text) - 1

            diffc = obj.find('difficult')
            difficult = 0 if diffc is None else int(diffc.text)

            catname = obj.find('name').text.lower().strip()
            w = x2 - x1
            h = y2 - y1
            ann = {
                'bbox': [x1, y1, w, h],
                'category_name': catname,
                'difficult': difficult,
                'weight': 1.0 - difficult,
            }
            cid = dset._alias_to_cat(ann.pop('category_name'))['id']
            dset.add_annotation(image_id=gid, category_id=cid, **ann)

    dset.dump(join(root, dset.tag + '.mscoco.json'), newlines=True)
    return dset


def _read_split_paths(devkit_dpath, split, year):
    """
    split = 'train'
    self = VOCDataset('test')
    year = 2007
    year = 2012
    """
    import glob
    import re
    split_idstrs = []
    data_dpath = join(devkit_dpath, 'VOC{}'.format(year))
    split_dpath = join(data_dpath, 'ImageSets', 'Main')
    pattern = join(split_dpath, '*_' + split + '.txt')
    for p in sorted(glob.glob(pattern)):
        rows = [list(re.split(' +', t)) for t in ub.readfrom(p).split('\n') if t]
        # code = -1 if the image does not contain the object
        # code = 1 if the image contains at least one instance
        # code = 0 if the image contains only hard instances of the object
        idstrs = [idstr for idstr, code in rows if int(code) == 1]
        split_idstrs.extend(idstrs)
    split_idstrs = sorted(set(split_idstrs))

    image_dpath = join(data_dpath, 'JPEGImages')
    annot_dpath = join(data_dpath, 'Annotations')
    gpaths = [join(image_dpath, '{}.jpg'.format(idstr))
              for idstr in split_idstrs]
    apaths = [join(annot_dpath, '{}.xml'.format(idstr))
              for idstr in split_idstrs]
    return gpaths, apaths


def ensure_voc_data(dpath=None, force=False, years=[2007, 2012]):
    """
    Download the Pascal VOC 2007 data if it does not already exist.

    Example:
        >>> # xdoctest: +REQUIRES(--download)
        >>> devkit_dpath = ensure_voc_data()
    """
    if dpath is None:
        dpath = ub.expandpath('~/data/VOC')
    devkit_dpath = join(dpath, 'VOCdevkit')
    # if force or not exists(devkit_dpath):
    ub.ensuredir(dpath)

    fpath1 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCdevkit_08-Jun-2007.tar', dpath=dpath)
    if force or not exists(join(dpath, 'VOCdevkit', 'VOCcode')):
        ub.cmd('tar xvf "{}" -C "{}"'.format(fpath1, dpath), verbout=1)

    if 2007 in years:
        # VOC 2007 train+validation data
        fpath2 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtrainval_06-Nov-2007.tar', dpath=dpath)
        if force or not exists(join(dpath, 'VOCdevkit', 'VOC2007', 'ImageSets', 'Main', 'bird_trainval.txt')):
            ub.cmd('tar xvf "{}" -C "{}"'.format(fpath2, dpath), verbout=1)

        # VOC 2007 test data
        fpath3 = ub.grabdata('http://host.robots.ox.ac.uk/pascal/VOC/voc2007/VOCtest_06-Nov-2007.tar', dpath=dpath)
        if force or not exists(join(dpath, 'VOCdevkit', 'VOC2007', 'ImageSets', 'Main', 'bird_test.txt')):
            ub.cmd('tar xvf "{}" -C "{}"'.format(fpath3, dpath), verbout=1)

    if 2012 in years:
        # VOC 2012 train+validation data
        fpath4 = ub.grabdata('https://pjreddie.com/media/files/VOCtrainval_11-May-2012.tar', dpath=dpath)
        if force or not exists(join(dpath, 'VOCdevkit', 'VOC2012', 'ImageSets', 'Main', 'bird_trainval.txt')):
            ub.cmd('tar xvf "{}" -C "{}"'.format(fpath4, dpath), verbout=1)
    return devkit_dpath


def main():
    convert_voc_to_coco()


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.data.grab_voc
    """
    main()



# -*- coding: utf-8 -*-
"""
Mostly deprecated in favor of kwplot
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import cv2
import itertools as it
import numpy as np
import six
import ubelt as ub
from os.path import join, dirname


def pandas_plot_matrix(df, rot=90, ax=None, grid=True, label=None,
                       zerodiag=False,
                       cmap='viridis', showvals=False, logscale=True):
    import matplotlib as mpl
    import copy
    from matplotlib import pyplot as plt
    import matplotlib.cm  # NOQA
    import kwplot
    if ax is None:
        fig = kwplot.figure(fnum=1, pnum=(1, 1, 1))
        fig.clear()
        ax = plt.gca()
    ax = plt.gca()
    values = df.values
    if zerodiag:
        values = values.copy()
        values = values - np.diag(np.diag(values))

    # aximg = ax.imshow(values, interpolation='none', cmap='viridis')
    if logscale:
        from matplotlib.colors import LogNorm
        vmin = df[df > 0].min().min()
        norm = LogNorm(vmin=vmin, vmax=values.max())
    else:
        norm = None

    cmap = copy.copy(mpl.cm.get_cmap(cmap))  # copy the default cmap
    cmap.set_bad((0, 0, 0))

    aximg = ax.matshow(values, interpolation='none', cmap=cmap, norm=norm)
    # aximg = ax.imshow(values, interpolation='none', cmap='viridis', norm=norm)

    # ax.imshow(values, interpolation='none', cmap='viridis')
    ax.grid(False)
    cax = plt.colorbar(aximg, ax=ax)
    if label is not None:
        cax.set_label(label)

    ax.set_xticks(list(range(len(df.index))))
    ax.set_xticklabels([lbl[0:100] for lbl in df.index])
    for lbl in ax.get_xticklabels():
        lbl.set_rotation(rot)
    for lbl in ax.get_xticklabels():
        lbl.set_horizontalalignment('center')

    ax.set_yticks(list(range(len(df.columns))))
    ax.set_yticklabels([lbl[0:100] for lbl in df.columns])
    for lbl in ax.get_yticklabels():
        lbl.set_horizontalalignment('right')
    for lbl in ax.get_yticklabels():
        lbl.set_verticalalignment('center')

    # Grid lines around the pixels
    if grid:
        offset = -.5
        xlim = [-.5, len(df.columns)]
        ylim = [-.5, len(df.index)]
        segments = []
        for x in range(ylim[1]):
            xdata = [x + offset, x + offset]
            ydata = ylim
            segment = list(zip(xdata, ydata))
            segments.append(segment)
        for y in range(xlim[1]):
            xdata = xlim
            ydata = [y + offset, y + offset]
            segment = list(zip(xdata, ydata))
            segments.append(segment)
        bingrid = mpl.collections.LineCollection(segments, color='w', linewidths=1)
        ax.add_collection(bingrid)

    if showvals:
        x_basis = np.arange(len(df.columns))
        y_basis = np.arange(len(df.index))
        x, y = np.meshgrid(x_basis, y_basis)

        for c, r in zip(x.flatten(), y.flatten()):
            val = df.iloc[r, c]
            ax.text(c, r, val, va='center', ha='center', color='white')
    return ax


def axes_extent(axs, pad=0.0):
    """
    Get the full extent of a group of axes, including axes labels, tick labels,
    and titles.
    """
    import matplotlib as mpl
    def axes_parts(ax):
        yield ax
        for label in ax.get_xticklabels():
            if label.get_text():
                yield label
        for label in ax.get_yticklabels():
            if label.get_text():
                yield label
        xlabel = ax.get_xaxis().get_label()
        ylabel = ax.get_yaxis().get_label()
        for label in (xlabel, ylabel, ax.title):
            if label.get_text():
                yield label

    items = it.chain.from_iterable(axes_parts(ax) for ax in axs)
    extents = [item.get_window_extent() for item in items]
    #mpl.transforms.Affine2D().scale(1.1)
    extent = mpl.transforms.Bbox.union(extents)
    extent = extent.expanded(1.0 + pad, 1.0 + pad)
    return extent


def extract_axes_extents(fig, combine=False, pad=0.0):
    """
    Extracts the extent of each axes item in inches. The main purpose of this
    is to set `bbox_inches` in `fig.savefig`, such that only the important data
    is visualized.

    Args:
        fig (Figure): the figure
        combine (bool): if True returns the union of each extent
        pad (float): additional padding around each axes

    Returns:
        matplotlib.transforms.Bbox or list of matplotlib.transforms.Bbox
    """
    # Make sure we draw the axes first so we can
    # extract positions from the text objects
    import matplotlib as mpl
    fig.canvas.draw()

    # Group axes that belong together
    atomic_axes = []
    seen_ = set([])
    for ax in fig.axes:
        if ax not in seen_:
            atomic_axes.append([ax])
            seen_.add(ax)

    dpi_scale_trans_inv = fig.dpi_scale_trans.inverted()
    axes_bboxes_ = [axes_extent(axs, pad) for axs in atomic_axes]
    axes_extents_ = [extent.transformed(dpi_scale_trans_inv) for extent in axes_bboxes_]
    # axes_extents_ = axes_bboxes_
    if combine:
        # Grab include extents of figure text as well
        # FIXME: This might break on OSX
        # http://stackoverflow.com/questions/22667224/bbox-backend
        renderer = fig.canvas.get_renderer()
        for mpl_text in fig.texts:
            bbox = mpl_text.get_window_extent(renderer=renderer)
            extent_ = bbox.expanded(1.0 + pad, 1.0 + pad)
            extent = extent_.transformed(dpi_scale_trans_inv)
            # extent = extent_
            axes_extents_.append(extent)
        axes_extents = mpl.transforms.Bbox.union(axes_extents_)
        # if True:
        #     axes_extents.x0 = 0
        #     # axes_extents.y1 = 0
    else:
        axes_extents = axes_extents_
    return axes_extents


def adjust_subplots(left=None, right=None, bottom=None, top=None, wspace=None,
                    hspace=None, fig=None):
    """
    Kwargs:
        left (float): left side of the subplots of the figure
        right (float): right side of the subplots of the figure
        bottom (float): bottom of the subplots of the figure
        top (float): top of the subplots of the figure
        wspace (float): width reserved for blank space between subplots
        hspace (float): height reserved for blank space between subplots
    """
    from matplotlib import pyplot as plt
    kwargs = dict(left=left, right=right, bottom=bottom, top=top,
                  wspace=wspace, hspace=hspace)
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if fig is None:
        fig = plt.gcf()
    subplotpars = fig.subplotpars
    adjust_dict = subplotpars.__dict__.copy()
    del adjust_dict['validate']
    adjust_dict.update(kwargs)
    fig.subplots_adjust(**adjust_dict)


def render_figure_to_image(fig, dpi=None, transparent=None, **savekw):
    """
    Saves a figure as an image in memory.

    Args:
        fig (matplotlib.figure.Figure): figure to save

        dpi (int or str, Optional):
            The resolution in dots per inch.  If *None* it will default to the
            value ``savefig.dpi`` in the matplotlibrc file.  If 'figure' it
            will set the dpi to be the value of the figure.

        transparent (bool):
            If *True*, the axes patches will all be transparent; the
            figure patch will also be transparent unless facecolor
            and/or edgecolor are specified via kwargs.

        **savekw: other keywords passed to `fig.savefig`. Valid keywords
            include: facecolor, edgecolor, orientation, papertype, format,
            pad_inches, frameon.

    Returns:
        np.ndarray: an image in BGR or BGRA format.

    Notes:
        Be sure to use `fig.set_size_inches` to an appropriate size before
        calling this function.
    """
    import io
    import cv2
    # import matplotlib as mpl
    # axes_extents = extract_axes_extents(fig)
    # extent = mpl.transforms.Bbox.union(axes_extents)
    extent = 'tight'  # mpl might do this correctly these days
    with io.BytesIO() as stream:
        # This call takes 23% - 15% of the time depending on settings
        fig.savefig(stream, bbox_inches=extent, dpi=dpi,
                    transparent=transparent, **savekw)
        # fig.savefig(stream, **savekw)
        stream.seek(0)
        data = np.fromstring(stream.getvalue(), dtype=np.uint8)
    im_bgra = cv2.imdecode(data, cv2.IMREAD_UNCHANGED)
    return im_bgra


def savefig2(fig, fpath, **kwargs):
    """
    Does a tight layout and saves the figure with transparency

    DEPRICATE
    """
    import matplotlib as mpl
    if 'transparent' not in kwargs:
        kwargs['transparent'] = True
    if 'extent' not in kwargs:
        axes_extents = extract_axes_extents(fig)
        extent = mpl.transforms.Bbox.union(axes_extents)
        kwargs['extent'] = extent
    fig.savefig(fpath, **kwargs)


def copy_figure_to_clipboard(fig):
    """
    References:
        https://stackoverflow.com/questions/17676373/python-matplotlib-pyqt-copy-image-to-clipboard
    """
    print('Copying figure %d to the clipboard' % fig.number)
    import matplotlib as mpl
    app = mpl.backends.backend_qt5.qApp
    QtGui = mpl.backends.backend_qt5.QtGui
    im_bgra = render_figure_to_image(fig, transparent=True)
    im_rgba = cv2.cvtColor(im_bgra, cv2.COLOR_BGRA2RGBA)
    im = im_rgba
    QImage = QtGui.QImage
    qim = QImage(im.data, im.shape[1], im.shape[0], im.strides[0], QImage.Format_RGBA8888)
    clipboard = app.clipboard()
    clipboard.setImage(qim)

    # size = fig.canvas.size()
    # width, height = size.width(), size.height()
    # qim = QtGui.QImage(fig.canvas.buffer_rgba(), width, height, QtGui.QImage.Format_ARGB32)

    # QtWidgets = mpl.backends.backend_qt5.QtWidgets
    # pixmap = QtWidgets.QWidget.grab(fig.canvas)
    # clipboard.setPixmap(pixmap)


def _get_axis_xy_width_height(ax=None, xaug=0, yaug=0, waug=0, haug=0):
    """ gets geometry of a subplot """
    from matplotlib import pyplot as plt
    if ax is None:
        ax = plt.gca()
    autoAxis = ax.axis()
    xy     = (autoAxis[0] + xaug, autoAxis[2] + yaug)
    width  = (autoAxis[1] - autoAxis[0]) + waug
    height = (autoAxis[3] - autoAxis[2]) + haug
    return xy, width, height


_LEGEND_LOCATION = {
    'upper right':  1,
    'upper left':   2,
    'lower left':   3,
    'lower right':  4,
    'right':        5,
    'center left':  6,
    'center right': 7,
    'lower center': 8,
    'upper center': 9,
    'center':      10,
}


_BASE_FNUM = 9001


def _save_requested(fpath_, save_parts):
    raise NotImplementedError('havent done this yet')
    # dpi = ub.argval('--dpi', type_=int, default=200)
    from os.path import expanduser
    from matplotlib import pyplot as plt
    dpi = 200
    fpath_ = expanduser(fpath_)
    print('Figure save was requested')
    # arg_dict = ut.get_arg_dict(prefix_list=['--', '-'],
    #                            type_hints={'t': list, 'a': list})
    arg_dict = {}
    # HACK
    arg_dict = {
        key: (val[0] if len(val) == 1 else '[' + ']['.join(val) + ']')
        if isinstance(val, list) else val
        for key, val in arg_dict.items()
    }
    fpath_ = fpath_.format(**arg_dict)
    fpath_ = fpath_.replace(' ', '').replace('\'', '').replace('"', '')
    dpath = ub.argval('--dpath', type_=str, default=None)
    if dpath is None:
        gotdpath = False
        dpath = '.'
    else:
        gotdpath = True

    fpath = join(dpath, fpath_)
    if not gotdpath:
        dpath = dirname(fpath_)
    print('dpath = %r' % (dpath,))

    fig = plt.gcf()
    fig.dpi = dpi

    fpath_strict = ub.expandpath(fpath)
    CLIP_WHITE = ub.argflag('--clipwhite')
    from netharn import util

    if save_parts:
        # TODO: call save_parts instead, but we still need to do the
        # special grouping.

        # Group axes that belong together
        atomic_axes = []
        seen_ = set([])
        for ax in fig.axes:
            div = _get_plotdat(ax, _DF2_DIVIDER_KEY, None)
            if div is not None:
                df2_div_axes = _get_plotdat_dict(ax).get('df2_div_axes', [])
                seen_.add(ax)
                seen_.update(set(df2_div_axes))
                atomic_axes.append([ax] + df2_div_axes)
                # TODO: pad these a bit
            else:
                if ax not in seen_:
                    atomic_axes.append([ax])
                    seen_.add(ax)

        hack_axes_group_row = ub.argflag('--grouprows')
        if hack_axes_group_row:
            groupid_list = []
            for axs in atomic_axes:
                for ax in axs:
                    groupid = ax.colNum
                groupid_list.append(groupid)

            groups = ub.group_items(atomic_axes, groupid_list)
            new_groups = list(map(ub.flatten, groups.values()))
            atomic_axes = new_groups
            #[[(ax.rowNum, ax.colNum) for ax in axs] for axs in atomic_axes]
            # save all rows of each column

        subpath_list = save_parts(fig=fig, fpath=fpath_strict,
                                  grouped_axes=atomic_axes, dpi=dpi)
        absfpath_ = subpath_list[-1]

        if CLIP_WHITE:
            for subpath in subpath_list:
                # remove white borders
                util.clipwhite_ondisk(subpath, subpath)
    else:
        savekw = {}
        # savekw['transparent'] = fpath.endswith('.png') and not noalpha
        savekw['transparent'] = ub.argflag('--alpha')
        savekw['dpi'] = dpi
        savekw['edgecolor'] = 'none'
        savekw['bbox_inches'] = extract_axes_extents(fig, combine=True)  # replaces need for clipwhite
        absfpath_ = ub.expandpath(fpath)
        fig.savefig(absfpath_, **savekw)

        if CLIP_WHITE:
            # remove white borders
            fpath_in = fpath_out = absfpath_
            util.clipwhite_ondisk(fpath_in, fpath_out)

    if ub.argflag(('--diskshow', '--ds')):
        # show what we wrote
        ub.startfile(absfpath_)


def save_parts(fig, fpath, grouped_axes=None, dpi=None):
    """
    FIXME: this works in mpl 2.0.0, but not 2.0.2

    Args:
        fig (?):
        fpath (str):  file path string
        dpi (None): (default = None)

    Returns:
        list: subpaths

    CommandLine:
        python -m draw_func2 save_parts

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> import kwplot
        >>> kwplot.autompl()
        >>> import matplotlib as mpl
        >>> import matplotlib.pyplot as plt
        >>> def testimg(fname):
        >>>     return plt.imread(mpl.cbook.get_sample_data(fname))
        >>> fnames = ['grace_hopper.png', 'ada.png'] * 4
        >>> fig = plt.figure(1)
        >>> for c, fname in enumerate(fnames, start=1):
        >>>     ax = fig.add_subplot(3, 4, c)
        >>>     ax.imshow(testimg(fname))
        >>>     ax.set_title(fname[0:3] + str(c))
        >>>     ax.set_xticks([])
        >>>     ax.set_yticks([])
        >>> ax = fig.add_subplot(3, 1, 3)
        >>> ax.plot(np.sin(np.linspace(0, np.pi * 2)))
        >>> ax.set_xlabel('xlabel')
        >>> ax.set_ylabel('ylabel')
        >>> ax.set_title('title')
        >>> fpath = 'test_save_parts.png'
        >>> adjust_subplots(fig=fig, wspace=.3, hspace=.3, top=.9)
        >>> subpaths = save_parts(fig, fpath, dpi=300)
        >>> fig.savefig(fpath)
        >>> ub.startfile(subpaths[0])
        >>> ub.startfile(fpath)
    """
    if dpi:
        # Need to set figure dpi before we draw
        fig.dpi = dpi
    # We need to draw the figure before calling get_window_extent
    # (or we can figure out how to set the renderer object)
    # if getattr(fig.canvas, 'renderer', None) is None:
    fig.canvas.draw()

    # Group axes that belong together
    if grouped_axes is None:
        grouped_axes = []
        for ax in fig.axes:
            grouped_axes.append([ax])

    subpaths = []
    _iter = enumerate(grouped_axes, start=0)
    _iter = ub.ProgIter(list(_iter), label='save subfig')
    for count, axs in _iter:
        subpath = ub.augpath(fpath, suffix=chr(count + 65))
        extent = axes_extent(axs).transformed(fig.dpi_scale_trans.inverted())
        savekw = {}
        savekw['transparent'] = ub.argflag('--alpha')
        if dpi is not None:
            savekw['dpi'] = dpi
        savekw['edgecolor'] = 'none'
        fig.savefig(subpath, bbox_inches=extent, **savekw)
        subpaths.append(subpath)
    return subpaths


_qtensured = False


def _current_ipython_session():
    """
    Returns a reference to the current IPython session, if one is running
    """
    try:
        __IPYTHON__
    except NameError:
        return None
    else:
        import IPython
        ipython = IPython.get_ipython()
        # if ipython is None we must have exited ipython at some point
        return ipython


def qtensure():
    """
    If you are in an IPython session, ensures that your backend is Qt.
    """
    global _qtensured
    if not _qtensured:
        ipython = _current_ipython_session()
        if ipython:
            import sys
            if 'PyQt4' in sys.modules:
                ipython.magic('pylab qt4 --no-import-all')
                _qtensured = True
            else:
                ipython.magic('pylab qt5 --no-import-all')
                _qtensured = True


def aggensure():
    """
    Ensures that you are in agg mode as long as IPython is not running

    This might help prevent errors in tmux like:
        qt.qpa.screen: QXcbConnection: Could not connect to display localhost:10.0
        Could not connect to any X display.
    """
    import matplotlib as mpl
    current_backend = mpl.get_backend()
    if current_backend != 'agg':
        ipython = _current_ipython_session()
        if not ipython:
            import kwplot
            kwplot.set_mpl_backend('agg')


def colorbar(scalars, colors, custom=False, lbl=None, ticklabels=None,
             float_format='%.2f', **kwargs):
    """
    adds a color bar next to the axes based on specific scalars

    Args:
        scalars (ndarray):
        colors (ndarray):
        custom (bool): use custom ticks

    Kwargs:
        See plt.colorbar

    Returns:
        cb : matplotlib colorbar object

    Ignore:
        >>> import kwplot
        >>> kwplot.autompl()
        >>> scalars = np.array([-1, -2, 1, 1, 2, 7, 10])
        >>> cmap_ = 'plasma'
        >>> logscale = False
        >>> custom = True
        >>> reverse_cmap = True
        >>> val2_customcolor  = {
        ...        -1: UNKNOWN_PURP,
        ...        -2: LIGHT_BLUE,
        ...    }
        >>> colors = scores_to_color(scalars, cmap_=cmap_, logscale=logscale, reverse_cmap=reverse_cmap, val2_customcolor=val2_customcolor)
        >>> colorbar(scalars, colors, custom=custom)
        >>> df2.present()
        >>> show_if_requested()

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> scalars = np.linspace(0, 1, 100)
        >>> cmap_ = 'plasma'
        >>> logscale = False
        >>> custom = False
        >>> reverse_cmap = False
        >>> colors = scores_to_color(scalars, cmap_=cmap_, logscale=logscale,
        >>>                          reverse_cmap=reverse_cmap)
        >>> colors = [lighten_rgb(c, .3) for c in colors]
        >>> colorbar(scalars, colors, custom=custom)
        >>> df2.present()
        >>> show_if_requested()
    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import matplotlib.cm  # NOQA
    assert len(scalars) == len(colors), 'scalars and colors must be corresponding'
    if len(scalars) == 0:
        return None
    # Parameters
    ax = plt.gca()
    divider = _ensure_divider(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)

    xy, width, height = _get_axis_xy_width_height(ax)
    #orientation = ['vertical', 'horizontal'][0]
    TICK_FONTSIZE = 8
    #
    # Create scalar mappable with cmap
    if custom:
        # FIXME: clean this code up and change the name custom
        # to be meaningful. It is more like: display unique colors
        unique_scalars, unique_idx = np.unique(scalars, return_index=True)
        unique_colors = np.array(colors)[unique_idx]
        #max_, min_ = unique_scalars.max(), unique_scalars.min()
        #extent_ = max_ - min_
        #bounds = np.linspace(min_, max_ + 1, extent_ + 2)
        listed_cmap = mpl.colors.ListedColormap(unique_colors)
        #norm = mpl.colors.BoundaryNorm(bounds, listed_cmap.N)
        #sm = mpl.cm.ScalarMappable(cmap=listed_cmap, norm=norm)
        sm = mpl.cm.ScalarMappable(cmap=listed_cmap)
        sm.set_array(np.linspace(0, 1, len(unique_scalars) + 1))
    else:
        sorted_scalars = sorted(scalars)
        listed_cmap = scores_to_cmap(scalars, colors)
        sm = plt.cm.ScalarMappable(cmap=listed_cmap)
        sm.set_array(sorted_scalars)
    # Use mapable object to create the colorbar
    #COLORBAR_SHRINK = .42  # 1
    #COLORBAR_PAD = .01  # 1
    #COLORBAR_ASPECT = np.abs(20 * height / (width))  # 1

    cb = plt.colorbar(sm, cax=cax, **kwargs)

    ## Add the colorbar to the correct label
    #axis = cb.ax.yaxis  # if orientation == 'horizontal' else cb.ax.yaxis
    #position = 'bottom' if orientation == 'horizontal' else 'right'
    #axis.set_ticks_position(position)

    # This line alone removes data
    # axis.set_ticks([0, .5, 1])
    if custom:
        ticks = np.linspace(0, 1, len(unique_scalars) + 1)
        if len(ticks) < 2:
            ticks += .5
        else:
            # SO HACKY
            ticks += (ticks[1] - ticks[0]) / 2

        if isinstance(unique_scalars, np.ndarray) and unique_scalars.dtype.kind == 'f':
            ticklabels = [float_format % scalar for scalar in unique_scalars]
        else:
            ticklabels = unique_scalars
        cb.set_ticks(ticks)  # tick locations
        cb.set_ticklabels(ticklabels)  # tick labels
    elif ticklabels is not None:
        ticks_ = cb.ax.get_yticks()
        mx = ticks_.max()
        mn = ticks_.min()
        ticks = np.linspace(mn, mx, len(ticklabels))
        cb.set_ticks(ticks)  # tick locations
        cb.set_ticklabels(ticklabels)
        #cb.ax.get_yticks()
        #cb.set_ticks(ticks)  # tick locations
        #cb.set_ticklabels(ticklabels)  # tick labels
    # _set_plotdat(cb.ax, 'viztype', 'colorbar-%s' % (lbl,))
    # _set_plotdat(cb.ax, 'sm', sm)
    # FIXME: Figure out how to make a maximum number of ticks
    # and to enforce them to be inside the data bounds
    cb.ax.tick_params(labelsize=TICK_FONTSIZE)
    # Sets current axis
    plt.sca(ax)
    if lbl is not None:
        cb.set_label(lbl)
    return cb


_DF2_DIVIDER_KEY = '_df2_divider'


def _get_plotdat(ax, key, default=None):
    """ returns internal property from a matplotlib axis """
    _plotdat = _get_plotdat_dict(ax)
    val = _plotdat.get(key, default)
    return val


def _set_plotdat(ax, key, val):
    """ sets internal property to a matplotlib axis """
    _plotdat = _get_plotdat_dict(ax)
    _plotdat[key] = val


def _del_plotdat(ax, key):
    """ sets internal property to a matplotlib axis """
    _plotdat = _get_plotdat_dict(ax)
    if key in _plotdat:
        del _plotdat[key]


def _get_plotdat_dict(ax):
    """ sets internal property to a matplotlib axis """
    if '_plotdat' not in ax.__dict__:
        ax.__dict__['_plotdat'] = {}
    plotdat_dict = ax.__dict__['_plotdat']
    return plotdat_dict


def _ensure_divider(ax):
    """ Returns previously constructed divider or creates one """
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = _get_plotdat(ax, _DF2_DIVIDER_KEY, None)
    if divider is None:
        divider = make_axes_locatable(ax)
        _set_plotdat(ax, _DF2_DIVIDER_KEY, divider)
        orig_append_axes = divider.append_axes
        def df2_append_axes(divider, position, size, pad=None, add_to_figure=True, **kwargs):
            """ override divider add axes to register the divided axes """
            div_axes = _get_plotdat(ax, 'df2_div_axes', [])
            new_ax = orig_append_axes(position, size, pad=pad, add_to_figure=add_to_figure, **kwargs)
            div_axes.append(new_ax)
            _set_plotdat(ax, 'df2_div_axes', div_axes)
            return new_ax
        new_method = df2_append_axes.__get__(divider, divider.__class__)
        setattr(divider, 'append_axes', new_method)
        # ut.inject_func_as_method(divider, df2_append_axes, 'append_axes', allow_override=True)
    return divider


def scores_to_cmap(scores, colors=None, cmap_='hot'):
    import matplotlib as mpl
    if colors is None:
        colors = scores_to_color(scores, cmap_=cmap_)
    scores = np.array(scores)
    colors = np.array(colors)
    sortx = scores.argsort()
    sorted_colors = colors[sortx]
    # Make a listed colormap and mappable object
    listed_cmap = mpl.colors.ListedColormap(sorted_colors)
    return listed_cmap


def scores_to_color(score_list, cmap_='hot', logscale=False, reverse_cmap=False,
                    custom=False, val2_customcolor=None, score_range=None,
                    cmap_range=(.1, .9)):
    """
    Other good colormaps are 'spectral', 'gist_rainbow', 'gist_ncar', 'Set1',
    'Set2', 'Accent'
    # TODO: plasma

    Args:
        score_list (list):
        cmap_ (str): defaults to hot
        logscale (bool):
        cmap_range (tuple): restricts to only a portion of the cmap to avoid extremes

    Returns:
        <class '_ast.ListComp'>

    Ignore:
        >>> ut.exec_funckw(scores_to_color, globals())
        >>> score_list = np.array([-1, -2, 1, 1, 2, 10])
        >>> # score_list = np.array([0, .1, .11, .12, .13, .8])
        >>> # score_list = np.linspace(0, 1, 100)
        >>> cmap_ = 'plasma'
        >>> colors = scores_to_color(score_list, cmap_)
        >>> imgRGB = kwarray.atleast_nd(np.array(colors)[:, 0:3], 3, tofront=True)
        >>> imgRGB = imgRGB.astype(np.float32)
        >>> imgBGR = kwimage.convert_colorspace(imgRGB, 'BGR', 'RGB')
        >>> imshow(imgBGR)
        >>> show_if_requested()

    Ignore:
        >>> score_list = np.array([-1, -2, 1, 1, 2, 10])
        >>> cmap_ = 'hot'
        >>> logscale = False
        >>> reverse_cmap = True
        >>> custom = True
        >>> val2_customcolor  = {
        ...        -1: UNKNOWN_PURP,
        ...        -2: LIGHT_BLUE,
        ...    }
    """
    import matplotlib.pyplot as plt
    assert len(score_list.shape) == 1, 'score must be 1d'
    if len(score_list) == 0:
        return []

    def apply_logscale(scores):
        scores = np.array(scores)
        above_zero = scores >= 0
        scores_ = scores.copy()
        scores_[above_zero] = scores_[above_zero] + 1
        scores_[~above_zero] = scores_[~above_zero] - 1
        scores_ = np.log2(scores_)
        return scores_

    if logscale:
        # Hack
        score_list = apply_logscale(score_list)
        #if loglogscale
        #score_list = np.log2(np.log2(score_list + 2) + 1)
    #if isinstance(cmap_, six.string_types):
    cmap = plt.get_cmap(cmap_)
    #else:
    #    cmap = cmap_
    if reverse_cmap:
        cmap = reverse_colormap(cmap)
    #if custom:
    #    base_colormap = cmap
    #    data = score_list
    #    cmap = customize_colormap(score_list, base_colormap)
    if score_range is None:
        min_ = score_list.min()
        max_ = score_list.max()
    else:
        min_ = score_range[0]
        max_ = score_range[1]
        if logscale:
            min_, max_ = apply_logscale([min_, max_])
    if cmap_range is None:
        cmap_scale_min, cmap_scale_max = 0., 1.
    else:
        cmap_scale_min, cmap_scale_max = cmap_range
    extent_ = max_ - min_
    if extent_ == 0:
        colors = [cmap(.5) for fx in range(len(score_list))]
    else:
        if False and logscale:
            # hack
            def score2_01(score):
                return np.log2(
                    1 + cmap_scale_min + cmap_scale_max *
                    (float(score) - min_) / (extent_))
            score_list = np.array(score_list)
            #rank_multiplier = score_list.argsort() / len(score_list)
            #normscore = np.array(list(map(score2_01, score_list))) * rank_multiplier
            normscore = np.array(list(map(score2_01, score_list)))
            colors =  list(map(cmap, normscore))
        else:
            def score2_01(score):
                return cmap_scale_min + cmap_scale_max * (float(score) - min_) / (extent_)
        colors = [cmap(score2_01(score)) for score in score_list]
        if val2_customcolor is not None:
            colors = [
                np.array(val2_customcolor.get(score, color))
                for color, score in zip(colors, score_list)]
    return colors


def interpolated_colormap(colors, resolution=64, space='lch-ab'):
    """
    Interpolates between colors in `space` to create a smooth listed colormap

    Args:
        colors (list or dict): list of colors or color objects and
            where in the map they should appear.

        resolution (int): number of discrete items in the colormap

        space (str): colorspace to interpolate in, using a CIE-LAB space will
            result in a perceptually uniform interpolation. HSV also works
            well.

    References:
        http://stackoverflow.com/questions/12073306/customize-colorbar-in-matplotlib

    CommandLine:
        python -m netharn.util.mplutil interpolated_colormap

    Example:
        >>> # DISABLE_DOCTEST
        >>> import kwplot
        >>> colors = [
        >>>     (0.0, kwplot.Color('green')),
        >>>     (0.5, kwplot.Color('gray')),
        >>>     (1.0, kwplot.Color('red')),
        >>> ]
        >>> space = 'lab'
        >>> #resolution = 16 + 1
        >>> resolution = 256 + 1
        >>> cmap = interpolated_colormap(colors, resolution, space)
        >>> # xdoc: +REQUIRES(--show)
        >>> import pylab
        >>> from matplotlib import pyplot as plt
        >>> a = np.linspace(0, 1, resolution).reshape(1, -1)
        >>> pylab.imshow(a, aspect='auto', cmap=cmap, interpolation='nearest')  # , origin="lower")
        >>> plt.grid(False)
        >>> show_if_requested()
    """
    import colorsys
    import matplotlib as mpl

    colors_inputs = colors

    if isinstance(colors_inputs, dict):
        colors_inputs = [(f, c) for f, c in sorted(colors_inputs.items())]
    else:
        if len(colors_inputs[0]) != 2:
            fracs = np.linspace(0, 1, len(colors_inputs))
            colors_inputs = list(zip(fracs, colors_inputs))

    # print('colors_inputs = {!r}'.format(colors_inputs))
    import kwplot
    colors = [kwplot.Color(c) for f, c in colors_inputs]
    fracs = [f for f, c in colors_inputs]

    basis = np.linspace(0, 1, resolution)
    fracs = np.array(fracs)
    indices = np.searchsorted(fracs, basis)
    indices = np.maximum(indices, 1)
    cpool = []

    from colormath import color_conversions
    # FIXME: need to ensure monkeypatch for networkx 2.0 in colormath
    # color_conversions._conversion_manager = color_conversions.GraphConversionManager()
    from colormath import color_objects
    def new_convertor(target_obj):
        source_obj = color_objects.sRGBColor
        def to_target(src_tup):
            src_tup = src_tup[0:3]
            src_co = source_obj(*src_tup)
            target_co = color_conversions.convert_color(src_co, target_obj)
            target_tup = target_co.get_value_tuple()
            return target_tup

        def from_target(target_tup):
            target_co = target_obj(*target_tup)
            src_co = color_conversions.convert_color(target_co, source_obj)
            src_tup = src_co.get_value_tuple()
            return src_tup
        return to_target, from_target

    def from_hsv(rgb):
        return colorsys.rgb_to_hsv(*rgb[0:3])

    def to_hsv(hsv):
        return colorsys.hsv_to_rgb(*hsv[0:3].tolist())

    classnames = {
        # 'AdobeRGBColor',
        # 'BaseRGBColor',
        'cmk': 'CMYColor',
        'cmyk': 'CMYKColor',
        'hsl': 'HSLColor',
        'hsv': 'HSVColor',
        'ipt': 'IPTColor',
        'lch-ab': 'LCHabColor',
        'lch-uv': 'LCHuvColor',
        'lab': 'LabColor',
        'luv': 'LuvColor',
        # 'SpectralColor',
        'xyz':  'XYZColor',
        # 'sRGBColor',
        'xyy': 'xyYColor'
    }

    conversions = {k: new_convertor(getattr(color_objects, v))
                   for k, v in classnames.items()}

    from_rgb, to_rgb = conversions['hsv']
    from_rgb, to_rgb = conversions['xyz']
    from_rgb, to_rgb = conversions['lch-uv']
    from_rgb, to_rgb = conversions['lch-ab']
    from_rgb, to_rgb = conversions[space]
    # from_rgb, to_rgb = conversions['lch']
    # from_rgb, to_rgb = conversions['lab']
    # from_rgb, to_rgb = conversions['lch-uv']

    for idx2, b in zip(indices, basis):
        idx1 = idx2 - 1
        f1 = fracs[idx1]
        f2 = fracs[idx2]

        c1 = colors[idx1].as01('rgb')
        c2 = colors[idx2].as01('rgb')
        # from_rgb, to_rgb = conversions['lch']
        h1 = np.array(from_rgb(c1))
        h2 = np.array(from_rgb(c2))
        alpha = (b - f1) / (f2 - f1)
        new_h = h1 * (1 - alpha) + h2 * (alpha)
        new_c = np.clip(to_rgb(new_h), 0, 1)
        # print('new_c = %r' % (new_c,))
        cpool.append(new_c)

    cpool = np.array(cpool)
    cmap = mpl.colors.ListedColormap(cpool, 'indexed')
    return cmap


def reverse_colormap(cmap):
    """
    References:
        http://nbviewer.ipython.org/github/kwinkunks/notebooks/blob/master/Matteo_colourmaps.ipynb
    """
    import matplotlib as mpl
    if isinstance(cmap,  mpl.colors.ListedColormap):
        return mpl.colors.ListedColormap(cmap.colors[::-1])
    else:
        reverse = []
        k = []
        for key, channel in six.iteritems(cmap._segmentdata):
            data = []
            for t in channel:
                data.append((1 - t[0], t[1], t[2]))
            k.append(key)
            reverse.append(sorted(data))
        cmap_reversed = mpl.colors.LinearSegmentedColormap(
            cmap.name + '_reversed', dict(zip(k, reverse)))
        return cmap_reversed


def draw_border(ax, color, lw=2, offset=None, adjust=True):
    'draws rectangle border around a subplot'
    if adjust:
        xy, width, height = _get_axis_xy_width_height(ax, -.7, -.2, 1, .4)
    else:
        xy, width, height = _get_axis_xy_width_height(ax)
    if offset is not None:
        xoff, yoff = offset
        xy = [xoff, yoff]
        height = - height - yoff
        width = width - xoff
    import matplotlib as mpl
    rect = mpl.patches.Rectangle(xy, width, height, lw=lw)
    rect = ax.add_patch(rect)
    rect.set_clip_on(False)
    rect.set_fill(False)
    import kwplot
    rect.set_edgecolor(kwplot.Color(color).as01('rgb'))
    return rect


def colorbar_image(domain, cmap='plasma', dpi=96, shape=(200, 20), transparent=False):
    """
    Notes:
        shape is approximate



    Ignore:
        domain = np.linspace(-30, 200)
        cmap='plasma'
        dpi = 80
        dsize = (20, 200)

        util.imwrite('foo.png', util.colorbar_image(np.arange(0, 1)), shape=(400, 80))

        import plottool as pt
        pt.qtensure()

        import matplotlib as mpl
        mpl.style.use('ggplot')
        util.imwrite('foo.png', util.colorbar_image(np.linspace(0, 1, 100), dpi=200, shape=(1000, 40), transparent=1))
        ub.startfile('foo.png')
    """
    import kwplot
    plt = kwplot.autoplt()

    fig = plt.figure(dpi=dpi)

    w, h = shape[1] / dpi, shape[0] / dpi
    # w, h = 1, 10
    fig.set_size_inches(w, h)

    ax = fig.add_subplot('111')

    sm = plt.cm.ScalarMappable(cmap=plt.get_cmap(cmap))
    sm.set_array(domain)

    plt.colorbar(sm, cax=ax)

    cb_img = render_figure_to_image(fig, dpi=dpi, transparent=transparent)

    plt.close(fig)

    return cb_img


def make_legend_img(classname_to_rgb, dpi=96, shape=(200, 200), mode='line',
                    transparent=False):
    """
    Makes an image of a categorical legend

    CommandLine:
        python -m netharn.util.mplutil make_legend_img

    Example:
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> classname_to_rgb = {
        >>>     'blue': kwplot.Color('blue').as01(),
        >>>     'red': kwplot.Color('red').as01(),
        >>> }
        >>> img = make_legend_img(classname_to_rgb)
        >>> # xdoctest: +REQUIRES(--show)
        >>> kwplot.autompl()
        >>> kwplot.imshow(img)
        >>> kwplot.show_if_requested()
    """
    import kwplot
    plt = kwplot.autoplt()

    def append_phantom_legend_label(label, color, type_='line', alpha=1.0, ax=None):
        if ax is None:
            ax = plt.gca()
        _phantom_legend_list = getattr(ax, '_phantom_legend_list', None)
        if _phantom_legend_list is None:
            _phantom_legend_list = []
            setattr(ax, '_phantom_legend_list', _phantom_legend_list)
        if type_ == 'line':
            phantom_actor = plt.Line2D((0, 0), (1, 1), color=color, label=label,
                                       alpha=alpha)
        else:
            phantom_actor = plt.Circle((0, 0), 1, fc=color, label=label,
                                       alpha=alpha)
        _phantom_legend_list.append(phantom_actor)

    fig = plt.figure(dpi=dpi)

    w, h = shape[1] / dpi, shape[0] / dpi
    fig.set_size_inches(w, h)

    ax = fig.add_subplot('111')
    for label, color in classname_to_rgb.items():
        append_phantom_legend_label(label, color, type_=mode, ax=ax)

    _phantom_legend_list = getattr(ax, '_phantom_legend_list', None)
    if _phantom_legend_list is None:
        _phantom_legend_list = []
        setattr(ax, '_phantom_legend_list', _phantom_legend_list)
    ax.legend(handles=_phantom_legend_list)
    ax.grid(False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    plt.axis('off')
    legend_img = render_figure_to_image(fig, dpi=dpi, transparent=transparent)
    plt.close(fig)
    return legend_img

if __name__ == '__main__':
    r"""
    CommandLine:
        python -m netharn.util.mplutil
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
DEPRICATE USE THE ONE IN YOLO2

Originally by EAVISE

Note: my alternative implementation of region_loss is in netharn/dev/ somewhere.
I probably should test if that is faster / more correct and them perhaps use
that instead.

"""
import math
import torch
import torch.nn as nn


class BaseLossWithCudaState(torch.nn.modules.loss._Loss):
    """
    Helper to keep track of if a loss module is in cpu or gpu mod
    """

    def __init__(self):
        super(BaseLossWithCudaState, self).__init__()
        self._iscuda = False
        self._device_num = None

    def cuda(self, device_num=None, **kwargs):
        self._iscuda = True
        self._device_num = device_num
        return super(BaseLossWithCudaState, self).cuda(device_num, **kwargs)

    def cpu(self):
        self._iscuda = False
        self._device_num = None
        return super(BaseLossWithCudaState, self).cpu()

    @property
    def is_cuda(self):
        return self._iscuda

    def get_device(self):
        if self._device_num is None:
            return torch.device('cpu')
        return self._device_num


class RegionLoss(BaseLossWithCudaState):
    """ Computes region loss from darknet network output and target annotation.

    Args:
        num_classes (int): number of categories
        anchors (list): 2D list representing anchor boxes (see :class:`lightnet.network.Darknet`)
        coord_scale (float): weight of bounding box coordinates
        noobject_scale (float): weight of regions without target boxes
        object_scale (float): weight of regions with target boxes
        class_scale (float): weight of categorical predictions
        thresh (float): minimum iou between a predicted box and ground truth for them to be considered matching
        seen (torch.Tensor): How many images the network has already been trained on.
    """

    def __init__(self, num_classes, anchors, reduction=32, seen=0, coord_scale=1.0, noobject_scale=1.0, object_scale=5.0, class_scale=1.0, thresh=0.6):
        super(RegionLoss, self).__init__()
        self.num_classes = num_classes
        self.num_anchors = len(anchors)
        self.anchor_step = len(anchors[0])
        self.anchors = torch.Tensor(anchors)
        self.reduction = reduction
        self.register_buffer('seen', torch.tensor(seen))

        self.coord_scale = coord_scale
        self.noobject_scale = noobject_scale
        self.object_scale = object_scale
        self.class_scale = class_scale
        self.thresh = thresh

        self.mse = nn.MSELoss(reduction='sum')
        self.clf = nn.CrossEntropyLoss(reduction='sum')

    # def extra_repr(self):
    #     repr_str = 'classes={self.num_classes}, reduction={self.reduction}, threshold={self.thresh}, seen={self.seen.item()}\n'
    #     repr_str += 'coord_scale={self.coord_scale}, object_scale={self.object_scale}, noobject_scale={self.noobject_scale}, class_scale={self.class_scale}\n'
    #     repr_str += 'anchors='
    #     for a in self.anchors:
    #         repr_str += '[{a[0]:.5g}, {a[1]:.5g}] '
    #     return repr_str

    def forward(self, output, target, seen=None):
        """ Compute Region loss.

        Args:
            output (torch.autograd.Variable): Output from the network
            target (brambox.boxes.annotations.Annotation or torch.Tensor): Brambox annotations or tensor containing the annotation targets (see :class:`lightnet.data.BramboxToTensor`)
            seen (int, optional): How many images the network has already been trained on; Default **Add batch_size to previous seen value**

        Note:
            The example below only shows this function working with a target tensor. |br|
            This loss function also works with a list of brambox annotations as target and will work the same.
            The added benefit of using brambox annotations is that this function will then also look at the ``ignore`` flag of the annotations
            and ignore detections that match with it. This allows you to have annotations that will not influence the loss in any way,
            as opposed to having them removed and counting them as false detections.

        Example:
            >>> # DISABLE_DOCTEST
            >>> from netharn.models.yolo2.region_loss2 import *
            >>> from netharn.models.yolo2.light_yolo import Yolo
            >>> torch.random.manual_seed(0)
            >>> network = Yolo(num_classes=2, conf_thresh=4e-2)
            >>> #_ = torch.random.manual_seed(0)
            >>> #network = ln.models.Yolo(num_classes=2, conf_thresh=4e-2)
            >>> #region_loss = ln.network.loss.RegionLoss(network.num_classes, network.anchors)
            >>> self = RegionLoss(network.num_classes, network.anchors)
            >>> Win, Hin = 96, 96
            >>> Wout, Hout = 1, 1
            >>> # true boxes for each item in the batch
            >>> # each box encodes class, x_center, y_center, width, and height
            >>> # coordinates are normalized in the range 0 to 1
            >>> # items in each batch are padded with dummy boxes with class_id=-1
            >>> target = torch.FloatTensor([
            ...     # boxes for batch item 1
            ...     [[0, 0.50, 0.50, 1.00, 1.00],
            ...      [1, 0.32, 0.42, 0.22, 0.12]],
            ...     # boxes for batch item 2 (it has no objects, note the pad!)
            ...     [[-1, 0, 0, 0, 0],
            ...      [-1, 0, 0, 0, 0]],
            ... ])
            >>> im_data = torch.randn(len(target), 3, Hin, Win, requires_grad=True)
            >>> output = network.forward(im_data)
            >>> loss = float(self(output, target))
            >>> print('loss = {loss:.2f}'.format(loss=loss))
            loss = 20.43
        """
        if isinstance(target, dict):
            target = target['target']

        # Get x,y,w,h,conf,cls
        if False:
            nB = output.data.size(0)
            nA = self.num_anchors
            nC = self.num_classes
            nH = output.data.size(2)
            nW = output.data.size(3)
            nPixels = nH * nW  # number of ouput grid cells

            device = output.device
            if seen is not None:
                self.seen = torch.tensor(seen)
            elif self.training:
                self.seen += nB

            output = output.view(nB, nA, -1, nPixels)
            coord = torch.zeros_like(output[:, :, :4])
            coord[:, :, :2] = output[:, :, :2].sigmoid()    # tx,ty
            coord[:, :, 2:4] = output[:, :, 2:4]            # tw,th
            conf = output[:, :, 4].sigmoid()
            if nC > 1:
                cls = output[:, :, 5:].contiguous().view(
                    nB * nA, nC, nPixels).transpose(1, 2).contiguous().view(-1, nC)

            # Create prediction boxes
            pred_boxes = torch.FloatTensor(nB * nA * nPixels, 4)
            lin_x = torch.linspace(0, nW - 1, nW).repeat(nH, 1).view(nPixels).to(device)
            lin_y = torch.linspace(0, nH - 1, nH).view(nH, 1).repeat(1, nW).view(nPixels).to(device)
            anchor_w = self.anchors[:, 0].contiguous().view(nA, 1).to(device)
            anchor_h = self.anchors[:, 1].contiguous().view(nA, 1).to(device)

            pred_boxes[:, 0] = (coord[:, :, 0].detach() + lin_x).view(-1)
            pred_boxes[:, 1] = (coord[:, :, 1].detach() + lin_y).view(-1)
            pred_boxes[:, 2] = (coord[:, :, 2].detach().exp() * anchor_w).view(-1)
            pred_boxes[:, 3] = (coord[:, :, 3].detach().exp() * anchor_h).view(-1)
            pred_boxes = pred_boxes.cpu()

        else:
            nB, nA, nC5, nH, nW = output.data.shape
            nC = self.num_classes
            assert nA == self.num_anchors
            assert nC5 == self.num_classes + 5
            output = output.view(nB, nA, 5 + nC, nH, nW)

            coord = torch.zeros_like(output[:, :, 0:4, :, :])
            coord[:, :, 0:2, :, :] = output[:, :, 0:2, :, :].sigmoid()  # tx,ty
            coord[:, :, 2:4, :, :] = output[:, :, 2:4, :, :]            # tw,th

            nPixels = nH * nW  # number of ouput grid cells

            device = output.device
            if seen is not None:
                self.seen = torch.tensor(seen)
            elif self.training:
                self.seen += nB

            conf = output[:, :, 4:5, :, :].sigmoid()
            if nC > 1:
                # Swaps the dimensions from [B, A, C, H, W] to be [B, A, H, W, C]
                cls = output[:, :, 5:, :, :].contiguous().view(
                    nB * nA, nC, nH * nW).transpose(1, 2).contiguous().view(
                        nB, nA, nH, nW, nC)

            pred_boxes = torch.empty(nB * nA * nH * nW, 4,
                                     dtype=torch.float32, device=device)
            lin_x = torch.linspace(0, nW - 1, nW).repeat(nH, 1).view(nPixels).to(device)
            lin_y = torch.linspace(0, nH - 1, nH).view(nH, 1).repeat(1, nW).view(nPixels).to(device)

            lin_x = torch.linspace(0, nW - 1, nW).repeat(nH, 1).to(device)
            lin_y = torch.linspace(0, nH - 1, nH).repeat(nW, 1).t().contiguous().to(device)

            anchor_w = self.anchors[:, 0].contiguous().view(nA, 1).view(1, nA, 1, 1, 1).to(device)
            anchor_h = self.anchors[:, 1].contiguous().view(nA, 1).view(1, nA, 1, 1, 1).to(device)

            pred_boxes[:, 0] = (coord[:, :, 0:1, :, :].detach() + lin_x).view(-1)
            pred_boxes[:, 1] = (coord[:, :, 1:2, :, :].detach() + lin_y).view(-1)
            pred_boxes[:, 2] = (coord[:, :, 2:3, :, :].detach().exp() * anchor_w).view(-1)
            pred_boxes[:, 3] = (coord[:, :, 3:4, :, :].detach().exp() * anchor_h).view(-1)

        # Get target values
        coord_mask, conf_mask, cls_mask, tcoord, tconf, tcls = self.build_targets(
            pred_boxes, target, nH, nW)
        coord_mask = coord_mask.expand_as(tcoord).to(device).sqrt()
        conf_mask = conf_mask.to(device).sqrt()
        tcoord = tcoord.to(device)
        tconf = tconf.to(device)

        if nC > 1:
            if False:
                tcls = tcls[cls_mask].view(-1).long().to(device)
                cls_mask = cls_mask.view(-1, 1).repeat(1, nC).to(device)
                cls = cls[cls_mask].view(-1, nC)
                cls_ = cls
            else:
                tcls = tcls[cls_mask].view(-1).long().to(device)
                cls_mask = cls_mask.view(-1, 1).repeat(1, nC).to(device)
                cls_ = cls.view(-1, nC)[cls_mask.view(-1, nC)].view(-1, nC)

            cls_ = cls_.view(-1, nC)
        coord_ = coord.view(coord_mask.shape)
        conf_ = conf.view(conf_mask.shape)

        # Compute losses
        self.loss_coord = self.coord_scale * (
            self.mse(coord_ * coord_mask, tcoord * coord_mask) / nB)
        self.loss_conf = self.mse(conf_ * conf_mask, tconf * conf_mask) / nB
        if nC > 1:
            self.loss_cls = self.class_scale * 2 * (self.clf(cls_, tcls) / nB)
            self.loss_tot = self.loss_coord + self.loss_conf + self.loss_cls
        else:
            self.loss_cls = None
            self.loss_tot = self.loss_coord + self.loss_conf

        return self.loss_tot

    def build_targets(self, pred_boxes, ground_truth, nH, nW):
        """ Compare prediction boxes and ground truths, convert ground truths to network output tensors """
        # Parameters
        nB = ground_truth.size(0)
        nA = self.num_anchors
        nAnchors = nA * nH * nW
        nPixels = nH * nW

        # Tensors
        conf_mask = torch.ones(
            nB, nA, nPixels, requires_grad=False) * self.noobject_scale
        coord_mask = torch.zeros(nB, nA, 1, nPixels, requires_grad=False)
        cls_mask = torch.zeros(nB, nA, nPixels, requires_grad=False).byte()
        tcoord = torch.zeros(nB, nA, 4, nPixels, requires_grad=False)
        tconf = torch.zeros(nB, nA, nPixels, requires_grad=False)
        tcls = torch.zeros(nB, nA, nPixels, requires_grad=False)

        if self.seen < 12800:
            coord_mask.fill_(1)
            # coord_mask.fill_(.01 / self.coord_scale)

            if self.anchor_step == 4:
                tcoord[:, :, 0] = self.anchors[:, 2].contiguous().view(
                    1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
                tcoord[:, :, 1] = self.anchors[:, 3].contiguous().view(
                    1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
            else:
                tcoord[:, :, 0].fill_(0.5)
                tcoord[:, :, 1].fill_(0.5)

        self.anchors = self.anchors.to(pred_boxes.device)
        if self.anchor_step == 4:
            anchors = self.anchors.clone()
            anchors[:, :2] = 0
        else:
            anchors = torch.cat(
                [torch.zeros_like(self.anchors), self.anchors], 1)

        for b in range(nB):
            gt = ground_truth[b][(ground_truth[b, :, 0] >= 0)[:, None].expand_as(ground_truth[b])].view(-1, 5)

            if gt.numel() > 0:     # No gt for this image
                # Build up tensors
                cur_pred_boxes = pred_boxes[b * nAnchors: (b + 1) * nAnchors]

                gt = gt[:, 1:]
                gt[:, ::2] *= nW
                gt[:, 1::2] *= nH

                # Set confidence mask of matching detections to 0
                iou_gt_pred = bbox_ious(gt, cur_pred_boxes)
                mask = (iou_gt_pred > self.thresh).sum(0) >= 1
                conf_mask[b][mask.view_as(conf_mask[b])] = 0

                # Find best anchor for each gt
                gt_wh = gt.clone()
                gt_wh[:, :2] = 0
                iou_gt_anchors = bbox_ious(gt_wh, anchors)
                _, best_anchors = iou_gt_anchors.max(1)

                # Set masks and target values for each gt
                gt_size = gt.size(0)
                for i in range(gt_size):
                    gi = min(nW - 1, max(0, int(gt[i, 0])))
                    gj = min(nH - 1, max(0, int(gt[i, 1])))
                    best_n = best_anchors[i]
                    iou = iou_gt_pred[i][best_n * nPixels + gj * nW + gi]

                    coord_mask[b][best_n][0][gj * nW + gi] = 2 - (gt[i, 2]  *  gt[i, 3]) / nPixels
                    cls_mask[b][best_n][gj * nW + gi] = 1
                    conf_mask[b][best_n][gj * nW + gi] = self.object_scale
                    tcoord[b][best_n][0][gj * nW + gi] = gt[i, 0] - gi
                    tcoord[b][best_n][1][gj * nW + gi] = gt[i, 1] - gj
                    tcoord[b][best_n][2][gj * nW + gi] = math.log(gt[i, 2] / self.anchors[best_n, 0])
                    tcoord[b][best_n][3][gj * nW  + gi] = math.log(gt[i, 3] / self.anchors[best_n, 1])
                    tconf[b][best_n][gj * nW + gi] = iou
                    tcls[b][best_n][gj * nW + gi] = ground_truth[b, i, 0]

        return coord_mask, conf_mask, cls_mask, tcoord, tconf, tcls


def bbox_ious(boxes1, boxes2):
    """ Compute IOU between all boxes from ``boxes1`` with all boxes from ``boxes2``.

    Args:
        boxes1 (torch.Tensor): List of bounding boxes
        boxes2 (torch.Tensor): List of bounding boxes

    Note:
        List format: [[xc, yc, w, h],...]
    """
    b1x1, b1y1 = (boxes1[:, :2] - (boxes1[:, 2:4] / 2)).split(1, 1)
    b1x2, b1y2 = (boxes1[:, :2] + (boxes1[:, 2:4] / 2)).split(1, 1)
    b2x1, b2y1 = (boxes2[:, :2] - (boxes2[:, 2:4] / 2)).split(1, 1)
    b2x2, b2y2 = (boxes2[:, :2] + (boxes2[:, 2:4] / 2)).split(1, 1)

    dx = (b1x2.min(b2x2.t()) - b1x1.max(b2x1.t())).clamp(min=0)
    dy = (b1y2.min(b2y2.t()) - b1y1.max(b2y1.t())).clamp(min=0)
    intersections = dx * dy

    areas1 = (b1x2 - b1x1) * (b1y2 - b1y1)
    areas2 = (b2x2 - b2x1) * (b2y2 - b2y1)
    unions = (areas1 + areas2.t()) - intersections

    return intersections / unions
